{Unit to implement big integers
Copyright (C) 2002-2010 Franco Milani
e-mail: nadir@solaris.polarhome.com or astlab@inwind.it
This software is part of the bigint package v. 1.8 (bigint.zip) available at:
http://www.polarhome.com/~franco/
http://www.polarhome.com:723/~nadir/software/
http://spazioinwind.libero.it/frm/software/
-----------------------------------------------------------------------------
This software is released in the hope that it will be useful, but on
an as-is basis. No warranty of any kind is offered, either expressed
or implied, including, without limitation, the implied warranties of
merchantability and fitness for a particular purpose.
You can use and modify this unit for personal purposes. You can also
redistribute it, but only in the original unmodified form and keeping
intact any part, included the above copyright notice. Possible changes
in the distributed versions will be made only by the author according
to hints, requests or bug reports.
You may distribute software items that link this unit or part of it
provided that you inform the recipients that this unit is employed.
-----------------------------------------------------------------------------
This unit contains routines to perform fast operations with big integers.
It provides only some support at low level. Error checking (division by zero,
overflow controls etc.) must be eventually implemented at higher level.
The natural numbers are defined in base 2**32 by a type nat, an array with
dimension dimnat whose elements are longwords. A nat N has the number of
its elements in N[0] and the elements starting from the least significant
in N[1], so it may be represented as:
N[1]+N[2]*base+N[3]*base**2+...+N[k]*base**(k-1), k=N[0], base=4294967296.
The signed integers are defined by a type zin, a record containing a nat and
a longint (the sign).
The maximum supported number of decimal digits can be set by replacing on the
source the maxdigits constant value as needed. At startup the unit converts
it automatically to a suitable corresponding array dimension by the formula
            dimnat = trunc((maxdigits-1)*ln(10)/(32*ln(2)))+2
The result exceeds at most by one the requested value.
Notice that the ansistrings converted by the procedures strton, ntostr,
strtoz and ztostr may have a leading minus if the number is negative, or no
leading characters if it is positive.
-----------------------------------------------------------------------------
References

Knuth's "Seminumerical Algorithms"
The Numerix library made by Michel Quercia
-----------------------------------------------------------------------------}
{$packrecords 1}
{$asmmode intel}
{$ifndef ver1_0}{$calling oldfpccall}{$endif}
{$H+}
UNIT Bigint18;

INTERFACE

CONST

maxdigits = 5000; {maximum supported digit number. May be changed as needed}

dimnat = trunc((maxdigits-1)*ln(10)/(32*ln(2)))+2;
psign = (dimnat+1)*4; {position of the sign in a zin}

TYPE

nat = array [0..dimnat] of longword;
zin = record
v: nat;
s: longint
end;

{************************  procedures and functions  *************************}

{* qwords *}

{number of binary digits (qbitsnum(q)=int(log2(q))+1)}
function qbitsnum (q: qword): longword;

{first non-zero bit, 0 if q = 0 or from 1 as first bit}
function qstartbit (q: qword): longword;

{true if bit nb set, false if zero. From 1 as first bit, up to qbitsnum(q)}
function qtestbit (q: qword; nb: longword): boolean;

{integer square root of a qword}
function sqrt64 (q: qword): longword;

{Jacobi Symbol: -1, 0 or 1 (p>0)}
function qjacobi (q,p: qword): longint;

{* longwords *}

{number of binary digits (bitsnum(n)=int(log2(n))+1)}
function bitsnum (n: longword): longword;

{first non-zero bit, 0 if n = 0 or from 1 as first bit}
function startbit (n: longword): longword;

{true if bit b set, false if zero. From 1 as first bit, up to bitsnum(n)}
function testbit (n: longword; b: longword): boolean;

{integer square root of a longword}
function sqrt32 (n: longword): longword;

{Jacobi Symbol: -1, 0 or 1 (m>0)}
function jacobi (n,m: longword): longint;

{* natural numbers *}

{- conversions from and to ansistrings -}

procedure strton (s: string; var a: nat);  {from ansistring to nat}
procedure ntostr (a: nat; var s: string);  {from nat to ansistring}

{- move -}

{moves a -> b}
procedure nmov (a: nat; var b: nat);

{- conversion from/to qword -}

{n(qword) -> a}
procedure qwrton (n: qword; var a: nat);

{a -> n(qword) (converts (a[1],a[2]) to qword)}
procedure ntoqwr (a: nat; var n: qword);

{- comparison functions -}

function ncmp (a,b: nat): longint;        {-1 if a < b, 0 if a = b, 1 if a > b}
function nequ (a,b: nat): boolean;             {true if a = b, false if a <> b}
function ndif (a,b: nat): boolean;             {true if a <> b, false if a = b}
function nlt (a,b: nat): boolean;              {true if a < b, false if a >= b}
function nlte (a,b: nat): boolean;             {true if a <= b, false if a > b}
function ngt (a,b: nat): boolean;              {true if a > b, false if a <= b}
function ngte (a,b: nat): boolean;             {true if a >= b, false if a < b}

{- bit scanning and digits -}

{number of binary digits, natbits(a)=int(log2(a))+1}
function natbits (a: nat): longword;

{number of decimal digits, ndsize(a)=int(log10(a))+1}
function ndsize (a: nat): longword;

{true if bit n set, false if zero. From 1 as first bit, up to natbits(a)}
function ntestbit (a: nat; n: longword): boolean;

{n exponent of the maximum power 2**n which divides a}
procedure nscan2 (a: nat; var n: longword);

{n exponent of the maximum power 2**n which divides a, a div 2**n -> b (a>0)}
procedure nsc2q (a: nat; var n: longword; var b: nat);

{n exponent of the maximum power 2**n which divides a, a div 2**n -> a (a>0)}
procedure nsc2qs (var a: nat; var n: longword);

{- bitwise logical operations -}

{a and b -> c}
procedure nand (a,b: nat; var c: nat);

{a or b -> c}
procedure nor (a,b: nat; var c: nat);

{a xor b -> c}
procedure nxor (a,b: nat; var c: nat);

{not a -> b}
procedure nnot (a: nat; var b: nat);

{- arithmetic operations -}

{a + b -> c}
procedure nadd (a,b: nat; var c: nat);

{a + b -> a}
procedure nadds (var a: nat; b: nat);

{a + n(longword) -> b}
procedure nadd_1 (a: nat; n: longword; var b: nat);

{a + n(longword) -> a}
procedure ninc (var a: nat; n: longword);

{a + 1 -> b}
procedure ninc1 (a: nat; var b: nat);

{a + 1 -> a}
procedure nincs (var a: nat);

{a - b -> c (a>=b)}
procedure nsub (a,b: nat; var c: nat);

{a - b -> a (a>=b)}
procedure nsubs (var a: nat; b: nat);

{a - b -> b (a>=b)}
procedure nsubv (a: nat; var b: nat);

{a - n(longword) -> b (a>=n)}
procedure nsub_1 (a: nat; n: longword; var b: nat);

{a - n(longword) -> a (a>=n)}
procedure ndec (var a: nat; n: longword);

{a - 1 -> b (a>=1)}
procedure ndec1 (a: nat; var b: nat);

{a - 1 -> a (a>=1)}
procedure ndecs (var a: nat);

{a * b -> c}
procedure nmul (a,b: nat; var c: nat);

{a * b -> a}
procedure nmuls (var a: nat; b: nat);

{a * n(longword) -> b}
procedure nmul_1 (a: nat; n: longword; var b: nat);

{a * n(longword) -> a}
procedure nmuls_1 (var a: nat; n: longword);

{a * a -> b}
procedure nsqr (a: nat; var b: nat);

{a * a -> a}
procedure nsqrs (var a: nat);

{a div b -> c, a mod b -> d (b>0)}
procedure ndiv (a,b: nat; var c,d: nat);

{a div n(longword) -> b, a mod n(longword) -> r (n>0)}
procedure ndiv_1 (a: nat; n: longword; var b: nat; var r: longword);

{a div b -> c (b>0)}
procedure nquo (a,b: nat; var c: nat);

{a div b -> a (b>0)}
procedure nquos (var a: nat; b: nat);

{a div n(longword) -> b (n>0)}
procedure nquo_1 (a: nat; n: longword; var b: nat);

{a div n(longword) -> a (n>0)}
procedure nquos_1 (var a: nat; n: longword);

{a mod b -> c (b>0)}
procedure nmod (a,b: nat; var c:nat);

{a mod b -> a (b>0)}
procedure nmods (var a: nat; b: nat);

{a mod n(longword) -> r(longword) (n>0)}
procedure nmod_1 (a: nat; n: longword; var r: longword);

{returns a mod n(longword) (n>0)}
function nmod0 (a: nat; n: longword): longword;

{- shifts -}

{a shr n(longword) -> b}
procedure nshr (a: nat; n: longword; var b: nat);

{a shr n(longword) -> a}
procedure nshrs (var a: nat; n: longword);

{a shr 1 -> a}
procedure nshr1 (var a: nat);

{a shl n(longword) -> b}
procedure nshl (a: nat; n: longword; var b: nat);

{a shl n(longword) -> a}
procedure nshls (var a: nat; n: longword);

{a shl 1 -> a}
procedure nshl1 (var a: nat);

{- roots -}

{square root of a -> b}
procedure nsqrt (a: nat; var b: nat);

{nth integer root of a -> b}
procedure nroot (a: nat; n: longword; var b: nat);

{- power -}

{a**n(longword) -> b}
procedure npow_1 (a: nat; n: longword; var b: nat);

{- exponentiation -}

{(a**b) mod c -> d (b>0 and c>0)}
procedure nmodexp (a,b,c: nat; var d: nat);

{2**n -> a}
procedure npw2 (var a: nat; n: longword);

{- composite operations -}

{(a+b) mod c -> d (c>0)}
procedure naddmod (a,b,c: nat; var d: nat);

{(a-b) mod c -> d (a>=b and c>0)}
procedure nsubmod (a,b,c: nat; var d: nat);

{(a*b) mod c -> d (c>0)}
procedure nmulmod (a,b,c: nat; var d: nat);

{(a*a) mod b -> c (b>0)}
procedure nsqrmod (a,b: nat; var c: nat);

{- greatest common divisor (gcd) and least common multiple (lcm) -}

{gcd(a,b) -> c (a>0 and b>0)}
procedure ngcd (a,b: nat; var c: nat);

{returns the g.c.d. of a(nat) and n(longword) (n>0)}
function ngcd0 (a: nat; n: longword): longword;

{lcm(a,b) -> c (a>0 and b>0)}
procedure nlcm (a,b: nat; var c: nat);

{- Jacobi symbol -}

{-1, 0 or 1 (b>0)}
function njacobi (a,b: nat): longint;

{- Fibonacci numbers -}

{n-th Fibonacci number -> a}
procedure fibn (n: longword; var a: nat);

{- binomial coefficient -}

{(n,k) -> a. Returns 0 if overflow}
procedure cbin (n,k: longword; var a: nat);

{- factorial -}

{n! -> a. Returns 0 if overflow}
procedure factr (n: longword; var a: nat);

{* signed integers *}

{- conversions from and to ansistrings -}

procedure strtoz (s: string; var a: zin);  {from ansistring to zin}
procedure ztostr (a: zin; var s: string);  {from zin to ansistring}

{- move -}

{moves a -> b}
procedure zmov (a: zin; var b: zin);

{- conversion from longint -}

{l(longint) -> a}
procedure ltoz (l: longint; var a: zin);

{- conversion from int64 -}

{n(int64) -> a}
procedure i64toz (n: int64; var a: zin);

{- comparison functions -}

function zcmp (a: zin; b: zin): longint;  {-1 if a < b, 0 if a = b, 1 if a > b}
function zequ (a: zin; b: zin): boolean;       {true if a = b, false if a < >b}
function zdif (a: zin; b: zin): boolean;       {true if a <> b, false if a = b}
function zlt (a: zin; b: zin): boolean;        {true if a < b, false if a >= b}
function zlte (a: zin; b: zin): boolean;       {true if a <= b, false if a > b}
function zgt (a: zin; b: zin): boolean;        {true if a > b, false if a <= b}
function zgte (a: zin; b: zin): boolean;       {true if a >= b, false if a < b}

{- arithmetic operations -}

{-a -> a}
procedure zchs (var a: zin);

{a + b -> c}
procedure zadd (a,b: zin; var c: zin);

{a + n(longword) -> b}
procedure zadd_1 (a: zin; n: longword; var b: zin);

{a - b -> c}
procedure zsub (a,b: zin; var c: zin);

{a - n(longword) -> b}
procedure zsub_1 (a: zin; n: longword; var b: zin);

{a * b -> c}
procedure zmul (a,b: zin; var c: zin);

{a * n(longword) -> b}
procedure zmul_1 (a: zin; n: longword; var b: zin);

{a * a -> b}
procedure zsqr (a: zin; var b: zin);

{- euclidean division (a = b*c + d, 0 <= d < |b|) -}

{euclidean quotient a div b -> c, modulo a mod b -> d (b<>0)}
procedure zediv (a,b: zin; var c,d: zin);

{euclidean quotient a div b -> c (b<>0)}
procedure zequo (a,b: zin; var c: zin);

{euclidean quotient a div n(longword) -> b (n>0)}
procedure zequo_1 (a: zin; n: longword; var b:zin);

{euclidean modulo a mod b -> c (b<>0)}
procedure zemod (a,b: zin; var c: zin);

{euclidean modulo a mod n(longword) -> r(longword) (n>0)}
procedure zemod_1 (a: zin; n: longword; var r: longword);

{- integer division (c = a/b, d = a - b*c) -}

{quotient a/b -> c, remainder d (b<>0)}
procedure zdiv (a,b: zin; var c,d: zin);

{quotient a/b -> c (b<>0)}
procedure zquo (a,b: zin; var c: zin);

{quotient a/n(longword) -> b (n>0)}
procedure zquo_1 (a: zin; n: longword; var b:zin);

{c = remainder of a/b (b<>0)}
procedure zrem (a,b: zin; var c: zin);

{b = remainder of a/n(longword) (n>0)}
procedure zrem_1 (a: zin; n: longword; var b: zin);

{returns |a| mod n(longword) (n>0)}
function zmod0 (a: zin; n: longword): longword;

{power a**n(longword) -> b}
procedure zpow_1 (a: zin; n: longword; var b: zin);

{- composite operations -}

{(a+b) mod c -> d (c<>0)}
procedure zaddmod (a,b,c: zin; var d: zin);

{(a-b) mod c -> d (c<>0)}
procedure zsubmod (a,b,c: zin; var d: zin);

{(a*b) mod c -> d (c<>0)}
procedure zmulmod (a,b,c: zin; var d: zin);

{(a*a) mod b -> c (b<>0)}
procedure zsqrmod (a,b: zin; var c: zin);

{exponentiation modulo (absolute values): (|a|**|b|) mod |c| -> d (b<>0, c<>0)}
procedure zmodexp (a,b,c: zin; var d: zin);

{greatest common divisor (absolute values): gcd(|a|,|b|) -> c (a<>0, b<>0)}
procedure zgcd (a,b: zin; var c: zin);

IMPLEMENTATION

VAR

natz1,natz2,natz3,natz4: zin;
ut,nt,qt,mp: nat;
dc,qc,qa,qb,qp: qword;
i,j,lz,la1,la2,pm,kz,cn,nc,iw,kv,ofs1,ofs2,ofs3,qn,ck,rc,dl,cl,el: longword;
reg_a,reg_b,reg_c,reg_q,reg_s,reg_u: longword;
mul0,loc_la,loc_lb,loc_a,loc_b: longword;
rs: longint;

{* qwords *}

function qbitsnum (q: qword): longword; assembler;
asm
                mov     ebx,[ebp+12]
                cmp     ebx,0
                je      @@1
                bsr     eax,ebx
                add     eax,33
                jmp     @@2
@@1:            mov     ebx,[ebp+8]
                cmp     ebx,0
                je      @@2
                bsr     eax,ebx
                inc     eax
@@2:
end;

function qstartbit (q: qword): longword; assembler;
asm
                bsf     eax,[ebp+8]
                jnz     @@1
                bsf     eax,[ebp+12]
                jz      @@2
                add     eax,32
@@1:            inc     eax
@@2:
end;

function qtestbit (q: qword; nb: longword): boolean; assembler;
asm
                mov     ebx,[ebp+16]
                dec     ebx
                cmp     ebx,32
                jb      @@1
                mov     eax,[ebp+12]
                jmp     @@2
@@1:            mov     eax,[ebp+8]
@@2:            bt      eax,ebx
                setc    al
end;

function sqrt64 (q: qword): longword;
begin
qn:=qbitsnum(q);
ck:=qn shr 1;
if ck<2 then dc:=4
else begin
if qn mod 2 = 0 then qc:=5 else qc:=7;
dec(ck,2);
dc:=qc shl ck
end;
repeat
qc:=q div dc;
inc(dc,qc);
dc:=dc shr 1;
until dc<=qc;
sqrt64:=dc
end;

function qjacobi (q,p: qword): longint;
begin
qa:=q mod p;
if qa=0 then qjacobi:=0
else begin
rs:=1;
qb:=p;
repeat
rc:=qstartbit(qa)-1;
qa:=qa shr rc;
if rc mod 2 = 1 then begin
rc:=qb and 7;
if (rc=3) or (rc=5) then rs:=-rs
end;
if qa=1 then break;
if (qb and 3 = 3) and (qa and 3 = 3) then rs:=-rs;
qp:=qa;
qa:=qb mod qa;
if qa=0 then begin
rs:=0;
break
end;
qb:=qp;
until false;
qjacobi:=rs
end
end;

{* longwords *}

function bitsnum (n: longword): longword; assembler;
asm
                bsr     eax,[ebp+8]
                jz      @@1
                inc     eax
@@1:
end;

function startbit (n: longword): longword; assembler;
asm
                bsf     eax,[ebp+8]
                jz      @@1
                inc     eax
@@1:
end;

function testbit (n: longword; b: longword): boolean; assembler;
asm
                mov     eax,[ebp+8]
                mov     ebx,[ebp+12]
                dec     ebx
                bt      eax,ebx
                setc    al
end;

function sqrt32 (n: longword): longword; assembler;
asm
                mov     eax,[ebp+8]
                cmp     eax,0
                je      @@2
                mov     edi,eax
                bsr     ecx,eax
                shr     ecx,1
                inc     ecx
                mov     ebx,1
                shl     ebx,cl
                shr     eax,cl
@@1:            add     eax,ebx
                mov     ebx,eax
                shr     ebx,1
                xor     edx,edx
                mov     eax,edi
                div     ebx
                cmp     ebx,eax
                ja      @@1
                mov     eax,ebx
@@2:
end;

function jacobi (n,m: longword): longint; assembler;
asm
                mov     eax,[ebp+8]
                mov     ebx,[ebp+12]
                xor     edx,edx
                div     ebx
                cmp     edx,0
                je      @@10
                mov     esi,1
@@20:           bsf     ecx,edx
                shr     edx,cl
                test    ecx,1
                jz      @@30
                mov     eax,ebx
                and     eax,7
                cmp     al,3
                je      @@40
                cmp     al,5
                jne     @@30
@@40:           neg     esi
@@30:           cmp     edx,1
                jne     @@50
                mov     eax,esi
                jmp     @@60
@@50:           mov     eax,ebx
                mov     edi,eax
                mov     ebx,edx
                xor     edx,edx
                div     ebx
                cmp     edx,0
                je      @@10
                and     edi,3
                cmp     edi,3
                jne     @@20
                mov     eax,ebx
                and     eax,3
                cmp     eax,3
                jne     @@20
                neg     esi
                jmp     @@20
@@10:           xor     eax,eax
@@60:
end;

{* natural numbers *}

procedure strton (s: string; var a: nat);
begin
a[0]:=1;
a[1]:=ord(s[1])-48;
for i:=2 to length(s) do begin
iw:=ord(s[i])-48;
nmuls_1(a,10);
ninc(a,iw)
end
end;

procedure ntostr (a: nat; var s: string);
begin
nmov(a,mp);
s:='';
repeat
ndiv_1(mp,10,mp,lz);
s:=chr(48+lz)+s;
until (mp[0]=1) and (mp[1]=0);
j:=length(s);
i:=1;
while (s[i]='0') and (i<j) do inc(i);
if i>1 then
if s[i]='0' then s:='0' else s:=copy(s,i,j-i+1)
end;

procedure nmov (a: nat; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi]
                mov     [esi],ecx
@@1:            mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
end;

procedure qwrton (n: qword; var a: nat); assembler;
asm
                mov     eax,[ebp+8]
                mov     ebx,[ebp+12]
                mov     esi,[ebp+16]
                mov     [esi+4],eax
                mov     dword ptr [esi],1
                cmp     ebx,0
                je      @@1
                mov     [esi+8],ebx
                inc     dword ptr [esi]
@@1:
end;

PROCEDURE ntoqwr (a: nat; var n: qword); assembler;
asm
                mov     esi,[ebp+8]
                mov     edi,[ebp+12]
                mov     eax,[esi+4]
                mov     [edi],eax
                mov     eax,0
                cmp     dword ptr [esi],1
                je      @@1
                mov     eax,[esi+8]
@@1:            mov     [edi+4],eax
end;

function ncmp (a: nat; b: nat): longint; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@1
                ja      @@2
@@4:            mov     eax,[edi+ecx*4]
                cmp     eax,[esi+ecx*4]
                jb      @@1
                ja      @@2
                dec     ecx
                jnz     @@4
@@3:            mov     eax,0
                jmp     @@5
@@1:            mov     eax,4294967295
                jmp     @@5
@@2:            mov     eax,1
@@5:
end;

function nequ (a: nat; b: nat): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                xor     eax,eax
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jne     @@5
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jne     @@5
                dec     ecx
                jnz     @@1
                inc     eax
@@5:
end;

function ndif (a: nat; b: nat): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     eax,1
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jne     @@5
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jne     @@5
                dec     ecx
                jnz     @@1
                dec     eax
@@5:
end;

function nlt (a: nat; b: nat): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     eax,1
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@5
                ja      @@4
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@5
                ja      @@4
                dec     ecx
                jnz     @@1
@@4:            dec     eax
@@5:
end;

function nlte (a: nat; b: nat): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                xor     eax,eax
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@4
                ja      @@5
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@4
                ja      @@5
                dec     ecx
                jnz     @@1
@@4:            inc     eax
@@5:
end;

function ngt (a: nat; b: nat): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     eax,1
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@4
                ja      @@5
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@4
                ja      @@5
                dec     ecx
                jnz     @@1
@@4:            dec     eax
@@5:
end;

function ngte (a: nat; b: nat): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                xor     eax,eax
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@5
                ja      @@4
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@5
                ja      @@4
                dec     ecx
                jnz     @@1
@@4:            inc     eax
@@5:
end;

function natbits (a: nat): longword; assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     eax,[edi+ecx*4]
                bsr     eax,eax
                jz      @@1
                inc     eax
                dec     ecx
                shl     ecx,5
                add     eax,ecx
@@1:
end;

function ndsize (a: nat): longword; assembler;
asm
                mov     esi,[ebp+8]
                lea     edi,natz1
                mov     ecx,[esi]
                mov     [edi],ecx
@@1:            mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     ebx,10
                xor     esi,esi
                mov     ecx,[edi]
@@2:            xor     edx,edx
@@3:            mov     eax,[edi+ecx*4]
                div     ebx
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@3
                inc     esi
                mov     ecx,[edi]
                cmp     dword ptr[edi+ecx*4],0
                jne     @@2
                dec     dword ptr [edi]
                jnz     @@2
                mov     eax,esi
end;

function ntestbit (a: nat; n: longword): boolean; assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[ebp+12]
                dec     ecx
                mov     ebx,ecx
                and     ebx,31
                shr     ecx,5
                inc     ecx
                mov     eax,[edi+ecx*4]
                bt      eax,ebx
                setc    al
end;

procedure nscan2 (a: nat; var n: longword); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                xor     edx,edx
                lea     esi,[esi+ecx*4+4]
                neg     ecx
                mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@4
                inc     ecx
                jz      @@5
                mov     edx,32
@@3:            mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@4
                add     edx,32
                inc     ecx
                jnz     @@3
                xor     edx,edx
                jmp     @@5
@@4:            bsf     ecx,eax
                add     edx,ecx
@@5:            mov     esi,[ebp+12]
                mov     [esi],edx
end;

procedure nsc2q (a: nat; var n: longword; var b: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                xor     edx,edx
                lea     esi,[esi+ecx*4+4]
                neg     ecx
                mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@10
                inc     ecx
                jz      @@11
                mov     edx,32
@@12:           mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@10
                add     edx,32
                inc     ecx
                jnz     @@12
                xor     edx,edx
                jmp     @@11
@@10:           bsf     ecx,eax
                add     edx,ecx
@@11:           mov     esi,[ebp+12]
                mov     [esi],edx
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ecx,edx
                mov     edx,[edi]
                cmp     ecx,0
                jne     @@13
                mov     [esi],edx
@@14:           mov     eax,[edi+edx*4]
                mov     [esi+edx*4],eax
                dec     edx
                jnz     @@14
                jmp     @@15
@@13:           cmp     ecx,32
                jb      @@16
                mov     eax,ecx
                shr     eax,5
                cmp     eax,edx
                jb      @@17
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@15
@@17:           and     ecx,31
                jnz     @@18
                mov     ecx,edx
                sub     ecx,eax
                mov     [esi],ecx
@@19:           mov     eax,[edi+edx*4]
                mov     [esi+ecx*4],eax
                dec     edx
                dec     ecx
                jnz     @@19
                jmp     @@15
@@18:           sub     edx,eax
                lea     edi,[edi+eax*4]
@@16:           mov     [esi],edx
                lea     edi,[edi+edx*4+4]
                lea     esi,[esi+edx*4]
                neg     edx
                mov     ebx,[edi+edx*4]
                inc     edx
                jnz     @@20
                shr     ebx,cl
                mov     [esi],ebx
                jmp     @@15
@@20:           mov     eax,[edi+edx*4]
                shrd    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                inc     edx
                jnz     @@20
                shr     ebx,cl
                mov     [esi],ebx
                mov     esi,[ebp+16]
                mov     ecx,[esi]
@@21:           cmp     dword ptr [esi+ecx*4],0
                jne     @@22
                dec     ecx
                jnz     @@21
                inc     ecx
@@22:           mov     [esi],ecx
@@15:
end;

procedure nsc2qs (var a: nat; var n: longword); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                xor     edx,edx
                lea     esi,[esi+ecx*4+4]
                neg     ecx
                mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@10
                inc     ecx
                jz      @@11
                mov     edx,32
@@12:           mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@10
                add     edx,32
                inc     ecx
                jnz     @@12
                xor     edx,edx
                jmp     @@11
@@10:           bsf     ecx,eax
                add     edx,ecx
@@11:           mov     esi,[ebp+12]
                mov     [esi],edx
                mov     edi,[ebp+8]
                mov     esi,edi
                mov     ecx,edx
                mov     edx,[edi]
                cmp     ecx,0
                je      @@13
@@14:           cmp     ecx,32
                jb      @@15
                mov     eax,ecx
                shr     eax,5
                cmp     eax,edx
                jb      @@16
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@13
@@16:           and     ecx,31
                jnz     @@17
                mov     ecx,edx
                sub     ecx,eax
                mov     [edi],ecx
                lea     esi,natz1
@@18:           mov     eax,[edi+edx*4]
                mov     [esi+ecx*4],eax
                dec     edx
                dec     ecx
                jnz     @@18
                mov     ecx,[edi]
@@19:           mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@19
                jmp     @@13
@@17:           sub     edx,eax
                lea     edi,[edi+eax*4]
@@15:           mov     [esi],edx
                lea     edi,[edi+edx*4+4]
                lea     esi,[esi+edx*4]
                neg     edx
                mov     ebx,[edi+edx*4]
                inc     edx
                jnz     @@20
                shr     ebx,cl
                mov     [esi],ebx
                jmp     @@13
@@20:           mov     eax,[edi+edx*4]
                shrd    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                inc     edx
                jnz     @@20
                shr     ebx,cl
                mov     [esi],ebx
                mov     edi,[ebp+8]
@@21:           mov     ecx,[edi]
                cmp     dword ptr [edi+ecx*4],0
                jne     @@13
                dec     ecx
                jz      @@13
@@22:           cmp     dword ptr [edi+ecx*4],0
                jne     @@23
                dec     ecx
                jnz     @@22
                inc     ecx
@@23:           mov     [edi],ecx
@@13:
end;

procedure nand (a: nat; b: nat; var c: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ebx,[ebp+16]
                mov     ecx,[edi]
                mov     edx,[esi]
                cmp     ecx,edx
                jbe     @@1
                xchg    ecx,edx
                xchg    esi,edi
@@1:            mov     edx,ecx
@@2:            mov     eax,[edi+ecx*4]
                and     eax,[esi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@2
@@3:            cmp     dword ptr [ebx+edx*4],0
                jne     @@4
                dec     edx
                jnz     @@3
                inc     edx
@@4:            mov     [ebx],edx
end;

procedure nor (a: nat; b: nat; var c: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ebx,[ebp+16]
                mov     ecx,[edi]
                mov     edx,[esi]
                cmp     ecx,edx
                je      @@3
                jb      @@1
                xchg    ecx,edx
                xchg    esi,edi
@@1:            push    edx
                push    ecx
                mov     eax,edx
                sub     eax,ecx
                mov     ecx,eax
@@2:            mov     eax,[esi+edx*4]
                mov     [ebx+edx*4],eax
                dec     edx
                dec     ecx
                jnz     @@2
                pop     ecx
                pop     edx
@@3:            mov     eax,[edi+ecx*4]
                or      eax,[esi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@3
@@4:            cmp     dword ptr [ebx+edx*4],0
                jne     @@5
                dec     edx
                jnz     @@4
                inc     edx
@@5:            mov     [ebx],edx
end;

procedure nxor (a: nat; b: nat; var c: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ebx,[ebp+16]
                mov     ecx,[edi]
                mov     edx,[esi]
                cmp     ecx,edx
                je      @@3
                jb      @@1
                xchg    ecx,edx
                xchg    esi,edi
@@1:            push    edx
                push    ecx
                mov     eax,edx
                sub     eax,ecx
                mov     ecx,eax
@@2:            mov     eax,[esi+edx*4]
                mov     [ebx+edx*4],eax
                dec     edx
                dec     ecx
                jnz     @@2
                pop     ecx
                pop     edx
@@3:            mov     eax,[edi+ecx*4]
                xor     eax,[esi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@3
@@4:            cmp     dword ptr [ebx+edx*4],0
                jne     @@5
                dec     edx
                jnz     @@4
                inc     edx
@@5:            mov     [ebx],edx
end;

procedure nnot (a: nat; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     edx,[edi]
                mov     eax,[edi+edx*4]
                bsr     ecx,eax
                cmp     ecx,31
                jne     @@1
                mov     ebx,4294967295
                jmp     @@2
@@1:            inc     ecx
                mov     ebx,1
                shl     ebx,cl
                dec     ebx
@@2:            xor     ebx,eax
                mov     [esi+edx*4],ebx
                mov     ecx,edx
                dec     ecx
                jz      @@5
                mov     ebx,4294967295
@@3:            mov     eax,[edi+ecx*4]
                xor     eax,ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@3
@@4:            cmp     dword ptr [esi+edx*4],0
                jne     @@5
                dec     edx
                jnz     @@4
                inc     edx
@@5:            mov     [esi],edx
end;

procedure nadd (a: nat; b: nat; var c: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@5
                xchg    edi,esi
                xchg    edx,ecx
@@5:            mov     ebx,[ebp+16]
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@6
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@7:            mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@7
                jnc     @@6
                test    edx,edx
                jz      @@8
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@9:            mov     eax,[esi+edx*4]
                inc     eax
                jne     @@10
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@9
@@8:            mov     dword ptr [ebx],1
                mov     ebx,[ebp+16]
                inc     dword ptr [ebx]
                jmp     @@11
@@6:            test    edx,edx
                jz      @@12
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@13:           mov     eax,[esi+edx*4]
@@10:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@13
@@12:           mov     [ebx],ecx
@@11:
end;

procedure nadds (var a: nat; b: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@5
                xchg    edi,esi
                xchg    edx,ecx
@@5:            lea     ebx,natz1
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@6
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@7:            mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@7
                jnc     @@6
                test    edx,edx
                jz      @@8
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@9:            mov     eax,[esi+edx*4]
                inc     eax
                jne     @@10
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@9
@@8:            mov     dword ptr [ebx],1
                lea     ebx,natz1
                inc     dword ptr [ebx]
                jmp     @@11
@@6:            test    edx,edx
                jz      @@12
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@13:           mov     eax,[esi+edx*4]
@@10:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@13
@@12:           mov     [ebx],ecx
@@11:           lea     ebx,natz1
                mov     edi,[ebp+8]
                mov     ecx,[ebx]
                mov     [edi],ecx
@@14:           mov     eax,[ebx+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@14
end;

procedure nadd_1 (a: nat; n: longword; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ecx,[edi]
                mov     edx,ecx
@@1:            mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     eax,[ebp+12]
                add     [esi+4],eax
                jc      @@5
                mov     [esi],edx
                jmp     @@10
@@5:            add     esi,4
                mov     ecx,edx
                dec     ecx
                jz      @@3
@@2:            add     esi,4
                inc     dword ptr [esi]
                jnz     @@4
                dec     ecx
                jnz     @@2
@@3:            mov     dword ptr[esi+4],1
                inc     edx
@@4:            mov     esi,[ebp+16]
                mov     [esi],edx
@@10:
end;

procedure ninc (var a: nat; n: longword); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     eax,[ebp+12]
                cmp     eax,0
                je      @@4
                add     esi,4
                add     [esi],eax
                jnc     @@4
                dec     ecx
                jz      @@3
@@2:            add     esi,4
                inc     dword ptr [esi]
                jnz     @@4
                dec     ecx
                jnz     @@2
@@3:            add     esi,4
                mov     dword ptr[esi],1
                mov     esi,[ebp+8]
                inc     dword ptr [esi]
@@4:
end;

procedure ninc1 (a: nat; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     esi,[ebp+12]
                mov     [esi],ecx
@@1:            mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     ecx,[edi]
@@2:            add     esi,4
                inc     dword ptr [esi]
                jnz     @@4
                dec     ecx
                jnz     @@2
                mov     dword ptr [esi+4],1
                mov     esi,[ebp+12]
                inc     dword ptr [esi]
@@4:
end;

procedure nincs (var a: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
@@1:            add     esi,4
                inc     dword ptr [esi]
                jnz     @@2
                dec     ecx
                jnz     @@1
                mov     dword ptr [esi+4],1
                mov     esi,[ebp+8]
                inc     dword ptr [esi]
@@2:
end;

procedure nsub (a: nat; b: nat; var c: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                mov     ebx,[ebp+16]
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@16
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@17:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@17
                jnb     @@16
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@18:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@19
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@18
                jmp     @@20
@@16:           test    edx,edx
                jz      @@20
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@21:           mov     eax,[esi+edx*4]
@@19:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@21
@@20:           mov     ebx,[ebp+16]
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jne     @@30
                dec     ecx
                jz      @@30
@@28:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@29
                dec     ecx
                jnz     @@28
                inc     ecx
@@29:           mov     [ebx],ecx
@@30:
end;

procedure nsubs (var a: nat; b: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                lea     ebx,natz1
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@16
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@17:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@17
                jnb     @@16
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@18:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@19
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@18
                jmp     @@20
@@16:           test    edx,edx
                jz      @@20
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@21:           mov     eax,[esi+edx*4]
@@19:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@21
@@20:           lea     ebx,natz1
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jne     @@30
                dec     ecx
                jz      @@30
@@28:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@29
                dec     ecx
                jnz     @@28
                inc     ecx
@@29:           mov     [ebx],ecx
@@30:           mov     edi,[ebp+8]
                mov     ecx,[ebx]
                mov     [edi],ecx
@@31:           mov     eax,[ebx+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@31
end;

procedure nsubv (a: nat; var b: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                sub     edx,ecx
                jecxz   @@16
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                neg     ecx
                clc
@@17:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [edi+ecx*4],eax
                inc     ecx
                jne     @@17
                jnb     @@16
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@18:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@19
                mov     [edi+edx*4],eax
                inc     edx
                jne     @@18
                jmp     @@20
@@16:           test    edx,edx
                jz      @@20
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@21:           mov     eax,[esi+edx*4]
@@19:           mov     [edi+edx*4],eax
                inc     edx
                jne     @@21
@@20:           mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                cmp     dword ptr[edi+ecx*4],0
                jne     @@30
                dec     ecx
                jz      @@29
@@28:           cmp     dword ptr[edi+ecx*4],0
                jne     @@30
                dec     ecx
                jnz     @@28
@@29:           inc     ecx
@@30:           mov     [edi],ecx
end;

procedure nsub_1 (a: nat; n: longword; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ecx,[edi]
                mov     edx,ecx
                mov     [esi],ecx
@@1:            mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     eax,[ebp+12]
                add     esi,4
                sub     [esi],eax
                jnc     @@4
                dec     edx
@@2:            add     esi,4
                sbb     dword ptr[esi],1
                jnc     @@3
                dec     edx
                jnz     @@2
                jmp     @@4
@@3:            jnz     @@4
                dec     edx
                jnz     @@4
                mov     esi,[ebp+16]
                dec     dword ptr [esi]
@@4:
end;

procedure ndec (var a: nat; n: longword); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     eax,[ebp+12]
                cmp     eax,0
                je      @@5
                add     esi,4
                sub     [esi],eax
                jnc     @@5
                dec     ecx
@@1:            add     esi,4
                sbb     dword ptr[esi],1
                jnc     @@2
                dec     ecx
                jnz     @@1
@@2:            cmp     dword ptr[esi],0
                jne     @@5
                mov     esi,[ebp+8]
                dec     dword ptr [esi]
@@5:
end;

procedure ndec1 (a: nat; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi]
                mov     ebx,ecx
                mov     [esi],ecx
@@1:            mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     eax,ebx
@@2:            add     esi,4
                sub     dword ptr [esi],1
                jnc     @@3
                dec     eax
                jnz     @@2
                jmp     @@4
@@3:            jnz     @@4
                dec     eax
                jnz     @@4
                dec     ebx
                jz      @@4
                mov     esi,[ebp+12]
                mov     [esi],ebx
@@4:
end;

procedure ndecs (var a: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     eax,ecx
@@1:            add     esi,4
                sub     dword ptr [esi],1
                jnc     @@2
                dec     ecx
                jnz     @@1
                jmp     @@3
@@2:            jnz     @@3
                dec     ecx
                jnz     @@3
                dec     eax
                jz      @@3
                mov     esi,[ebp+8]
                mov     [esi],eax
@@3:
end;

procedure nmul (a: nat; b: nat; var c: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                mov     edx,[edi]
                mov     ebx,[ebp+16]
                cmp     ecx,edx
                jae     @@5
                xchg    edi,esi
                xchg    edx,ecx
@@5:            mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@8:            mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@8
                mov     [ebx],edx
                dec     loc_lb
                jz      @@10
@@9:            add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@11:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@11
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@9
@@10:           mov     esi,[ebp+16]
                mov     ecx,pm
@@20:           cmp     dword ptr[esi+ecx*4],0
                jne     @@14
                dec     ecx
                jnz     @@20
                inc     ecx
@@14:           mov     [esi],ecx
end;

procedure nmuls (var a: nat; b: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                mov     edx,[edi]
                lea     ebx,natz1
                cmp     ecx,edx
                jae     @@5
                xchg    edi,esi
                xchg    edx,ecx
@@5:            mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@8:            mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@8
                mov     [ebx],edx
                dec     loc_lb
                jz      @@10
@@9:            add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@11:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@11
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@9
@@10:           lea     esi,natz1
                mov     edi,[ebp+8]
                mov     ecx,pm
@@14:           cmp     dword ptr[esi+ecx*4],0
                jne     @@12
                dec     ecx
                jnz     @@14
                inc     ecx
@@12:           mov     [edi],ecx
@@20:           mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@20
end;

procedure nmul_1 (a: nat; n: longword; var b: nat); assembler;
asm
                mov     esi,[ebp+16]
                mov     eax,[ebp+12]
                cmp     eax,0
                jne     @@1
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@10
@@1:            mov     pm,eax
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     [esi],ecx
                add     edi,4
                add     esi,4
                xor     ebx,ebx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                neg     ecx
@@2:            mov     eax,[edi+ecx*4]
                mul     pm
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jne     @@2
                cmp     ebx,0
                je      @@10
                mov     esi,[ebp+16]
                inc     dword ptr [esi]
                mov     ecx,[esi]
                mov     [esi+ecx*4],ebx
@@10:
end;

procedure nmuls_1 (var a: nat; n: longword); assembler;
asm
                mov     esi,[ebp+8]
                mov     edi,[ebp+12]
                cmp     edi,0
                jne     @@1
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@10
@@1:            mov     ecx,[esi]
                xor     ebx,ebx
                lea     esi,[esi+ecx*4+4]
                neg     ecx
@@2:            mov     eax,[esi+ecx*4]
                mul     edi
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jnz     @@2
                cmp     ebx,0
                je      @@10
                mov     esi,[ebp+8]
                inc     dword ptr [esi]
                mov     ecx,[esi]
                mov     [esi+ecx*4],ebx
@@10:
end;

procedure nsqr (a: nat; var b: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                mov     ebx,[ebp+12]
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@8:            mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@8
                mov     [ebx],edx
                dec     loc_lb
                jz      @@10
@@9:            add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@11:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@11
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@9
@@10:           mov     esi,[ebp+12]
                mov     ecx,pm
@@20:           cmp     dword ptr[esi+ecx*4],0
                jne     @@33
                dec     ecx
                jnz     @@20
                inc     ecx
@@33:           mov     [esi],ecx
end;

procedure nsqrs (var a: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz2
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@8:            mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@8
                mov     [ebx],edx
                dec     loc_lb
                jz      @@10
@@9:            add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@11:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@11
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@9
@@10:           lea     esi,natz2
                mov     edi,[ebp+8]
                mov     ecx,pm
@@20:           cmp     dword ptr[esi+ecx*4],0
                jne     @@27
                dec     ecx
                jnz     @@20
                inc     ecx
@@27:           mov     [edi],ecx
@@17:           mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@17
end;

procedure ndiv (a: nat; b: nat; var c: nat; var d: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@11
                mov     edi,[ebp+16]
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                mov     esi,[ebp+16]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@15
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@16
                dec     ecx
                jnz     @@16
                inc     ecx
@@16:           mov     [esi],ecx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                mov     esi,[ebp+16]
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@17:           sub     esi,4
                cmp     edx,reg_b
                je      @@18
                div     reg_b
                jmp     @@19
@@18:           add     edx,eax
                mov     eax,4294967295
                jnc     @@19
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@20
                mov     ebx,[edi]
@@20:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@21
@@19:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@22
                mov     ebx,[edi]
@@22:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@23
                ja      @@24
                test    eax,eax
                jz      @@23
@@24:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@23
                ja      @@24
                test    eax,eax
                jnz     @@24
@@23:           neg     eax
                adc     edx,0
                neg     edx
@@21:           popf
                jnz     @@17
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@25
                mov     dword ptr [edi],1
                jmp     @@26
@@25:           mov     [edi+8],edx
                mov     dword ptr [edi],2
@@26:           mov     edi,[ebp+16]
                mov     ecx,la1
@@27:           cmp     dword ptr [edi+ecx*4],0
                jne     @@28
                dec     ecx
                jnz     @@27
                inc     ecx
@@28:           mov     [edi],ecx
                jmp     @@12
@@13:           mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
                mov     ebx,[ebp+16]
@@29:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@29
@@30:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@31
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@32
@@31:           mov     ebx,[ebp+16]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@33
@@32:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@34:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@35
                div     reg_a
                mov     reg_q,eax
                jmp     @@36
@@35:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@37
@@36:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@37
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@37
                dec     reg_q
@@37:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@38:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@38
                cmp     [esi],ebx
                je      @@39
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@40:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@40
@@39:           mov     ebx,[ebp+16]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@34
@@33:           mov     ecx,la1
@@41:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@42
                dec     ecx
                jnz     @@41
                inc     ecx
@@42:           mov     [ebx],ecx
@@43:           mov     esi,[ebp+20]
                mov     ecx,la2
@@44:           cmp     dword ptr[esi+ecx*4],0
                jne     @@45
                dec     ecx
                jnz     @@44
                inc     ecx
@@45:           mov     [esi],ecx
@@12:
end;

procedure ndiv_1 (a: nat; n: longword; var b: nat; var r: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     ebx,[ebp+20]
                mov     [ebx],edx
                mov     ecx,[edi]
                cmp     dword ptr[esi+ecx*4],0
                jne     @@2
                dec     ecx
                jnz     @@2
                inc     ecx
@@2:            mov     [esi],ecx
end;

procedure nquo (a: nat; b: nat; var c: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
@@11:           mov     edi,[ebp+16]
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                mov     esi,[ebp+16]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@15
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@16
                dec     ecx
                jnz     @@16
                inc     ecx
@@16:           mov     [esi],ecx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                mov     esi,[ebp+16]
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@17:           sub     esi,4
                cmp     edx,reg_b
                je      @@18
                div     reg_b
                jmp     @@19
@@18:           add     edx,eax
                mov     eax,4294967295
                jnc     @@19
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@20
                mov     ebx,[edi]
@@20:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@21
@@19:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@22
                mov     ebx,[edi]
@@22:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@23
                ja      @@24
                test    eax,eax
                jz      @@23
@@24:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@23
                ja      @@24
                test    eax,eax
                jnz     @@24
@@23:           neg     eax
                adc     edx,0
                neg     edx
@@21:           popf
                jnz     @@17
                mov     edi,[ebp+16]
                mov     ecx,la1
@@25:           cmp     dword ptr [edi+ecx*4],0
                jne     @@26
                dec     ecx
                jnz     @@25
                inc     ecx
@@26:           mov     [edi],ecx
                jmp     @@12
@@13:           mov     edx,la1
                lea     esi,natz1
                mov     ebx,[ebp+16]
@@27:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@27
@@28:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@29
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@30
@@29:           mov     ebx,[ebp+16]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@31
@@30:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@32:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@33
                div     reg_a
                mov     reg_q,eax
                jmp     @@34
@@33:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@35
@@34:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@35
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@35
                dec     reg_q
@@35:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@36:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@36
                cmp     [esi],ebx
                je      @@37
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@38:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@38
@@37:           mov     ebx,[ebp+16]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@32
@@31:           mov     ecx,la1
@@39:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@40
                dec     ecx
                jnz     @@39
                inc     ecx
@@40:           mov [ebx],ecx
@@12:
end;

procedure nquos (var a: nat; b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@11
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@12
                je      @@13
                mov     ebx,[esi+4]
                xor     edx,edx
@@14:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@14
                mov     ecx,la1
                cmp     dword ptr[edi+ecx*4],0
                jne     @@15
                dec     ecx
                jnz     @@15
                inc     ecx
@@15:           mov     [edi],ecx
                jmp     @@11
@@13:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                lea     esi,natz1
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@16:           sub     esi,4
                cmp     edx,reg_b
                je      @@17
                div     reg_b
                jmp     @@18
@@17:           add     edx,eax
                mov     eax,4294967295
                jnc     @@18
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@19
                mov     ebx,[edi]
@@19:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@20
@@18:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@21
                mov     ebx,[edi]
@@21:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@22
                ja      @@23
                test    eax,eax
                jz      @@22
@@23:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@22
                ja      @@23
                test    eax,eax
                jnz     @@23
@@22:           neg     eax
                adc     edx,0
                neg     edx
@@20:           popf
                jnz     @@16
                lea     edi,natz1
                mov     esi,[ebp+8]
                mov     ecx,la1
@@24:           cmp     dword ptr [edi+ecx*4],0
                jne     @@25
                dec     ecx
                jnz     @@24
                inc     ecx
@@25:           mov     [esi],ecx
@@26:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@26
                jmp     @@11
@@12:           mov     edx,la1
                lea     esi,natz1
@@27:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [edi+ecx*4],0
                dec     ecx
                jnz     @@27
@@28:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@29
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@30
@@29:           mov     ebx,[ebp+8]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@31
@@30:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@32:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@33
                div     reg_a
                mov     reg_q,eax
                jmp     @@34
@@33:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@35
@@34:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@35
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@35
                dec     reg_q
@@35:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@36:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@36
                cmp     [esi],ebx
                je      @@37
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@38:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@38
@@37:           mov     ebx,[ebp+8]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@32
@@31:           mov     ecx,la1
@@39:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@40
                dec     ecx
                jnz     @@39
                inc     ecx
@@40:           mov [ebx],ecx
@@11:
end;

procedure nquo_1 (a: nat; n: longword; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     esi,[ebp+16]
                mov     ecx,[edi]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     ecx,[edi]
                cmp     dword ptr[esi+ecx*4],0
                jne     @@5
                dec     ecx
                jnz     @@5
                inc     ecx
@@5:            mov     [esi],ecx
end;

procedure nquos_1 (var a: nat; n: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     ecx,[edi]
                cmp     dword ptr[edi+ecx*4],0
                jne     @@5
                dec     dword ptr [edi]
                jnz     @@5
                inc     dword ptr [edi]
@@5:
end;

procedure nmod (a: nat; b: nat; var c: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@11
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@15
                mov     ebx,[ebp+16]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@16:           cmp     edx,reg_b
                je      @@17
                div     reg_b
                jmp     @@18
@@17:           add     edx,eax
                mov     eax,4294967295
                jnc     @@18
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@19
                mov     ebx,[edi]
@@19:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@20
@@18:           mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@21
                mov     ebx,[edi]
@@21:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@22
                ja      @@23
                test    eax,eax
                jz      @@22
@@23:           dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@22
                ja      @@23
                test    eax,eax
                jnz     @@23
@@22:           neg     eax
                adc     edx,0
                neg     edx
@@20:           popf
                jnz     @@16
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+16]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@24
                mov     dword ptr [edi],1
                jmp     @@12
@@24:           mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@12
@@13:           mov     edx,la1
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@25:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@25
@@26:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@27
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@28
@@27:           test    edx,edx
                jz      @@29
@@28:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@30:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@31
                div     reg_a
                mov     reg_q,eax
                jmp     @@32
@@31:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@33
@@32:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@33
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@33
                dec     reg_q
@@33:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@34:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@34
                cmp     [esi],ebx
                je      @@35
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@36:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@36
@@35:           dec     reg_c
                jne     @@30
@@29:           mov     esi,[ebp+16]
                mov     ecx,la2
@@37:           cmp     dword ptr[esi+ecx*4],0
                jne     @@38
                dec     ecx
                jnz     @@37
                inc     ecx
@@38:           mov     [esi],ecx
@@12:
end;

procedure nmods (var a: nat; b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                ja      @@10
                mov     la2,edx
                cmp     edx,2
                ja      @@11
                je      @@12
                mov     ebx,[esi+4]
                xor     edx,edx
@@13:           mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@13
                mov     dword ptr [edi],1
                mov     [edi+4],edx
                jmp     @@10
@@12:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@14:           cmp     edx,reg_b
                je      @@15
                div     reg_b
                jmp     @@16
@@15:           add     edx,eax
                mov     eax,4294967295
                jnc     @@16
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@17
                mov     ebx,[edi]
@@17:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@18
@@16:           mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@19
                mov     ebx,[edi]
@@19:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@20
                ja      @@21
                test    eax,eax
                jz      @@20
@@21:           dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@20
                ja      @@21
                test    eax,eax
                jnz     @@21
@@20:           neg     eax
                adc     edx,0
                neg     edx
@@18:           popf
                jnz     @@14
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+8]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@22
                mov     dword ptr [edi],1
                jmp     @@10
@@22:           mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@10
@@11:           mov     edx,la1
@@23:           lea     esi,[edi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@24
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@25
@@24:           test    edx,edx
                jz      @@26
@@25:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@27:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@28
                div     reg_a
                mov     reg_q,eax
                jmp     @@29
@@28:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@30
@@29:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@30
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@30
                dec     reg_q
@@30:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@31:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@31
                cmp     [esi],ebx
                je      @@32
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@33:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@33
@@32:           dec     reg_c
                jne     @@27
@@26:           mov     edi,[ebp+8]
                mov     ecx,la2
@@34:           cmp     dword ptr[edi+ecx*4],0
                jne     @@35
                dec     ecx
                jnz     @@34
                inc     ecx
@@35:           mov     [edi],ecx
@@10:
end;

procedure nmod_1 (a: nat; n: longword; var r: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                mov     esi,[ebp+16]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1
                mov     [esi],edx
end;

function nmod0 (a: nat; n: longword): longword; assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1
                mov     eax,edx
end;

procedure nshr (a: nat; n: longword; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     edx,[edi]
                mov     ecx,[ebp+12]
                cmp     ecx,0
                jne     @@10
                mov     [esi],edx
@@11:           mov     eax,[edi+edx*4]
                mov     [esi+edx*4],eax
                dec     edx
                jnz     @@11
                jmp     @@12
@@10:           cmp     ecx,32
                jb      @@13
                mov     eax,ecx
                shr     eax,5
                cmp     eax,edx
                jb      @@14
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@12
@@14:           and     ecx,31
                jnz     @@15
                mov     ecx,edx
                sub     ecx,eax
                mov     [esi],ecx
@@16:           mov     eax,[edi+edx*4]
                mov     [esi+ecx*4],eax
                dec     edx
                dec     ecx
                jnz     @@16
                jmp     @@12
@@15:           sub     edx,eax
                lea     edi,[edi+eax*4]
@@13:           mov     [esi],edx
                lea     edi,[edi+edx*4+4]
                lea     esi,[esi+edx*4]
                neg     edx
                mov     ebx,[edi+edx*4]
                inc     edx
                jnz     @@17
                shr     ebx,cl
                mov     [esi],ebx
                jmp     @@12
@@17:           mov     eax,[edi+edx*4]
                shrd    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                inc     edx
                jnz     @@17
                shr     ebx,cl
                mov     [esi],ebx
                mov     esi,[ebp+16]
                mov     ecx,[esi]
@@18:           cmp     dword ptr [esi+ecx*4],0
                jne     @@19
                dec     ecx
                jnz     @@18
                inc     ecx
@@19:           mov     [esi],ecx
@@12:
end;

procedure nshrs (var a: nat; n: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,edi
                mov     edx,[edi]
                mov     ecx,[ebp+12]
                cmp     ecx,0
                je      @@10
@@11:           cmp     ecx,32
                jb      @@12
                mov     eax,ecx
                shr     eax,5
                cmp     eax,edx
                jb      @@13
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@10
@@13:           and     ecx,31
                jnz     @@14
                mov     ecx,edx
                sub     ecx,eax
                mov     [edi],ecx
                lea     esi,natz1
@@15:           mov     eax,[edi+edx*4]
                mov     [esi+ecx*4],eax
                dec     edx
                dec     ecx
                jnz     @@15
                mov     ecx,[edi]
@@16:           mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@16
                jmp     @@10
@@14:           sub     edx,eax
                lea     edi,[edi+eax*4]
@@12:           mov     [esi],edx
                lea     edi,[edi+edx*4+4]
                lea     esi,[esi+edx*4]
                neg     edx
                mov     ebx,[edi+edx*4]
                inc     edx
                jnz     @@17
                shr     ebx,cl
                mov     [esi],ebx
                jmp     @@10
@@17:           mov     eax,[edi+edx*4]
                shrd    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                inc     edx
                jnz     @@17
                shr     ebx,cl
                mov     [esi],ebx
                mov     edi,[ebp+8]
@@18:           mov     ecx,[edi]
                cmp     dword ptr [edi+ecx*4],0
                jne     @@10
                dec     ecx
                jz      @@10
@@19:           cmp     dword ptr [edi+ecx*4],0
                jne     @@20
                dec     ecx
                jnz     @@19
                inc     ecx
@@20:           mov     [edi],ecx
@@10:
end;

procedure nshr1 (var a: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                clc
@@1:            rcr     dword ptr [edi+ecx*4],1
                dec     ecx
                jnz     @@1
                mov     ecx,[edi]
                cmp     dword ptr [edi+ecx*4],0
                jne     @@2
                dec     dword ptr [edi]
                jnz     @@2
                inc     dword ptr [edi]
@@2:
end;

procedure nshl (a: nat; n: longword; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     edx,[edi]
                mov     ecx,[ebp+12]
                cmp     edx,1
                jne     @@10
                cmp     dword ptr [edi+4],0
                jne     @@10
                mov     [esi],edx
                mov     dword ptr [esi+4],0
                jmp     @@11
@@10:           cmp     ecx,0
                jne     @@12
                mov     [esi],edx
@@13:           mov     eax,[edi+edx*4]
                mov     [esi+edx*4],eax
                dec     edx
                jnz     @@13
                jmp     @@11
@@12:           mov     eax,ecx
                shr     eax,5
                and     ecx,31
                jnz     @@14
                mov     ecx,edx
                add     edx,eax
                mov     [esi],edx
@@15:           mov     ebx,[edi+ecx*4]
                mov     [esi+edx*4],ebx
                dec     edx
                dec     ecx
                jnz     @@15
@@16:           mov     dword ptr [esi+edx*4],0
                dec     edx
                jnz     @@16
                jmp     @@11
@@14:           mov     kz,eax
                mov     [esi],edx
                lea     esi,[esi+4]
                xor     ebx,ebx
@@17:           mov     eax,[edi+edx*4]
                shld    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                dec     edx
                jnz     @@17
                shl     ebx,cl
                mov     [esi],ebx
                mov     esi,[ebp+16]
                mov     ebx,[esi]
                cmp     dword ptr [esi+ebx*4+4],0
                je      @@18
                inc     dword ptr [esi]
@@18:           cmp     kz,0
                je      @@11
                mov     ecx,[esi]
                mov     ebx,ecx
                add     ebx,kz
                mov     [esi],ebx
@@19:           mov     eax,[esi+ecx*4]
                mov     [esi+ebx*4],eax
                dec     ebx
                dec     ecx
                jnz     @@19
@@20:           mov     dword ptr [esi+ebx*4],0
                dec     ebx
                jnz     @@20
@@11:
end;

procedure nshls (var a: nat; n: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     edx,[edi]
                cmp     edx,1
                jne     @@100
                cmp     dword ptr [edi+4],0
                je      @@110
@@100:          mov     ecx,[ebp+12]
                cmp     ecx,0
                je      @@110
@@120:          mov     eax,ecx
                shr     eax,5
                and     ecx,31
                jnz     @@130
                mov     ecx,edx
                add     edx,eax
                mov     [edi],edx
@@140:          mov     eax,[edi+ecx*4]
                mov     [edi+edx*4],eax
                dec     edx
                dec     ecx
                jnz     @@140
@@150:          mov     dword ptr [edi+edx*4],0
                dec     edx
                jnz     @@150
                jmp     @@110
@@130:          mov     kz,eax
                lea     esi,[edi+4]
                xor     ebx,ebx
@@160:          mov     eax,[edi+edx*4]
                shld    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                dec     edx
                jnz     @@160
                shl     ebx,cl
                mov     [esi],ebx
                mov     ebx,[edi]
                cmp     dword ptr [edi+ebx*4+4],0
                je      @@170
                inc     dword ptr [edi]
@@170:          cmp     kz,0
                je      @@110
                mov     ecx,[edi]
                mov     ebx,ecx
                add     ebx,kz
                mov     [edi],ebx
@@180:          mov     eax,[edi+ecx*4]
                mov     [edi+ebx*4],eax
                dec     ebx
                dec     ecx
                jnz     @@180
@@190:          mov     dword ptr [edi+ebx*4],0
                dec     ebx
                jnz     @@190
@@110:
end;

procedure nshl1 (var a: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                xor     ebx,ebx
                clc
@@1:            inc     ebx
                rcl     dword ptr [edi+ebx*4],1
                dec     ecx
                jnz     @@1
                jnc     @@2
                inc     ebx
                mov     dword ptr [edi+ebx*4],1
                inc     dword ptr [edi]
@@2:
end;

procedure nsqrt (a: nat; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                cmp     dword ptr [edi],1
                jne     @@100
                cmp     dword ptr [edi+4],1
                ja      @@100
                mov     dword ptr [esi],1
                mov     eax,[edi+4]
                mov     [esi+4],eax
                jmp     @@110
@@100:          mov     ecx,[edi]
                mov     eax,[edi+ecx*4]
                cmp     eax,0
                je      @@120
                bsr     eax,eax
                inc     eax
@@120:          dec     ecx
                shl     ecx,5
                add     eax,ecx
                mov     edi,eax
                shr     eax,1
                cmp     eax,2
                jae     @@130
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],4
                jmp     @@140
@@130:          dec     eax
                dec     eax
                lea     esi,natz4
                mov     ebx,32
                xor     edx,edx
                div     ebx
                mov     ecx,eax
                inc     eax
                mov     [esi],eax
                jecxz   @@150
@@160:          mov     dword ptr[esi+ecx*4],0
                dec     ecx
                jnz     @@160
@@150:          mov     ebx,1
                mov     ecx,edx
                shl     ebx,cl
                mov     [esi+eax*4],ebx
                mov     eax,edi
                and     eax,1
                shl     eax,1
                add     eax,5
                mov     esi,[ebp+12]
                mov     pm,eax
                lea     edi,natz4
                mov     ecx,[edi]
                mov     [esi],ecx
                add     edi,4
                add     esi,4
                xor     ebx,ebx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                neg     ecx
@@170:          mov     eax,[edi+ecx*4]
                mul     pm
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jne     @@170
                cmp     ebx,0
                je      @@140
                mov     esi,[ebp+12]
                inc     dword ptr [esi]
                mov     ecx,[esi]
                mov     [esi+ecx*4],ebx
@@140:          mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@180
                lea     edi,natz4
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@190
@@180:          mov     la2,edx
                cmp     edx,2
                ja      @@200
                je      @@210
                mov     ebx,[esi+4]
                lea     esi,natz4
                xor     edx,edx
@@220:          mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@220
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@230
                dec     ecx
                jnz     @@230
                inc     ecx
@@230:          mov     [esi],ecx
                jmp     @@190
@@210:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                lea     esi,natz4
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@240:          sub     esi,4
                cmp     edx,reg_b
                je      @@250
                div     reg_b
                jmp     @@260
@@250:          add     edx,eax
                mov     eax,4294967295
                jnc     @@260
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@270
                mov     ebx,[edi]
@@270:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@280
@@260:          mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@290
                mov     ebx,[edi]
@@290:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@300
                ja      @@310
                test    eax,eax
                jz      @@300
@@310:          dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@300
                ja      @@310
                test    eax,eax
                jnz     @@310
@@300:          neg     eax
                adc     edx,0
                neg     edx
@@280:          popf
                jnz     @@240
                lea     edi,natz4
                mov     ecx,la1
@@320:          cmp     dword ptr [edi+ecx*4],0
                jne     @@330
                dec     ecx
                jnz     @@320
                inc     ecx
@@330:          mov     [edi],ecx
                jmp     @@190
@@200:          mov     edx,la1
                lea     esi,natz1
                lea     ebx,natz4
@@340:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@340
@@350:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@360
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@370
@@360:          lea     ebx,natz4
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@380
@@370:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@390:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@400
                div     reg_a
                mov     reg_q,eax
                jmp     @@410
@@400:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@420
@@410:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@420
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@420
                dec     reg_q
@@420:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@430:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@430
                cmp     [esi],ebx
                je      @@440
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@450:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@450
@@440:          lea     ebx,natz4
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@390
@@380:          mov     ecx,la1
@@460:          cmp     dword ptr[ebx+ecx*4],0
                jne     @@470
                dec     ecx
                jnz     @@460
                inc     ecx
@@470:          mov     [ebx],ecx
@@190:          mov     esi,[ebp+12]
                mov     edx,[esi]
                add     esi,4
                lea     edi,natz4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@480
                xchg    edi,esi
                xchg    edx,ecx
@@480:          mov     ebx,[ebp+12]
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@490
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@500:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@500
                jnc     @@490
                test    edx,edx
                jz      @@510
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@520:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@530
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@520
@@510:          mov     dword ptr [ebx],1
                mov     ebx,[ebp+12]
                inc     dword ptr [ebx]
                jmp     @@540
@@490:          test    edx,edx
                jz      @@550
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@560:          mov     eax,[esi+edx*4]
@@530:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@560
@@550:          mov     [ebx],ecx
@@540:          mov     edi,[ebp+12]
                mov     ecx,[edi]
                clc
@@570:          rcr     dword ptr [edi+ecx*4],1
                dec     ecx
                jnz     @@570
                mov     ecx,[edi]
                cmp     dword ptr [edi+ecx*4],0
                jne     @@580
                dec     dword ptr [edi]
                jnz     @@580
                inc     dword ptr [edi]
@@580:          mov     edi,[ebp+12]
                lea     esi,natz4
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@110
                ja      @@140
@@590:          mov     eax,[edi+ecx*4]
                cmp     eax,[esi+ecx*4]
                jb      @@110
                ja      @@140
                dec     ecx
                jnz     @@590
@@110:
end;

procedure nroot (a: nat; n: longword; var b: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                cmp     ebx,2
                jae     @@110
                mov     esi,[ebp+16]
                mov     ecx,[edi]
                mov     [esi],ecx
@@105:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@105
                jmp     @@9900
@@110:          cmp     dword ptr [edi],1
                jne     @@107
                cmp     dword ptr [edi+4],2
                jae     @@107
                mov     esi,[ebp+16]
                mov     dword ptr [esi],1
                mov     eax,[edi+4]
                mov     [esi+4],eax
                jmp     @@9900
@@107:          mov     cn,ebx
                lea     esi,UT
                mov     dword ptr [esi],1
                mov     edx,[edi]
                mov     eax,[edi+edx*4]
                bsr     eax,eax
                inc     eax
                dec     edx
                shl     edx,5
                add     eax,edx
                xor     edx,edx
                div     ebx
                cmp     eax,2
                jae     @@120
                mov     dword ptr [esi+4],4
                jmp     @@2500
@@120:          mov     ecx,eax
                sub     ecx,2
                mov     eax,edx
                shl     eax,2
                xor     edx,edx
                div     ebx
                add     eax,5
                mov     [esi+4],eax
                cmp     ecx,0
                je      @@2500
                mov     edx,[esi]
@@2000:         mov     eax,ecx
                shr     eax,5
                and     ecx,31
                jnz     @@2010
                mov     ecx,edx
                add     edx,eax
                mov     [esi],edx
@@2020:         mov     eax,[esi+ecx*4]
                mov     [esi+edx*4],eax
                dec     edx
                dec     ecx
                jnz     @@2020
@@2030:         mov     dword ptr [esi+edx*4],0
                dec     edx
                jnz     @@2030
                jmp     @@2500
@@2010:         mov     kz,eax
                lea     edi,[esi+4]
                xor     ebx,ebx
@@2050:         mov     eax,[esi+edx*4]
                shld    ebx,eax,cl
                mov     [edi+edx*4],ebx
                mov     ebx,eax
                dec     edx
                jnz     @@2050
                shl     ebx,cl
                mov     [edi],ebx
                mov     ebx,[esi]
                cmp     dword ptr [esi+ebx*4+4],0
                je      @@2060
                inc     dword ptr [esi]
@@2060:         cmp     kz,0
                je      @@2500
                mov     ecx,[esi]
                mov     ebx,ecx
                add     ebx,kz
                mov     [esi],ebx
@@2070:         mov     eax,[esi+ecx*4]
                mov     [esi+ebx*4],eax
                dec     ebx
                dec     ecx
                jnz     @@2070
@@2080:         mov     dword ptr [esi+ebx*4],0
                dec     ebx
                jnz     @@2080
@@2500:         mov     eax,[ebp+12]
                dec     eax
                mov     cn,eax
                lea     edi,UT
                mov     esi,[ebp+16]
@@2800:         mov     ecx,[edi]
                mov     [esi],ecx
@@2810:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@2810
                lea     esi,QT
                mov     edx,cn
                dec     edx
@@3100:         lea     edi,UT
                lea     ebx,natz3
                mov     ecx,[edi]
                mov     [esi],ecx
                mov     [ebx],ecx
@@3120:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@3120
                mov     kz,edx
@@3130:         test    kz,1
                jz      @@3140
                lea     esi,QT
                mov     ecx,[esi]
                lea     edi,natz3
                mov     edx,[edi]
                lea     ebx,natz1
                cmp     ecx,edx
                jae     @@3150
                xchg    edi,esi
                xchg    edx,ecx
@@3150:         mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@3160:         mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@3160
                mov     [ebx],edx
                dec     loc_lb
                jz      @@3170
@@3180:         add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@3190:         mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@3190
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@3180
@@3170:         lea     esi,natz1
                lea     edi,QT
                mov     ecx,pm
@@3200:         cmp     dword ptr[esi+ecx*4],0
                jne     @@3210
                dec     ecx
                jnz     @@3200
                inc     ecx
@@3210:         mov     [edi],ecx
@@3220:         mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@3220
@@3140:         shr     kz,1
                jz      @@3110
                lea     esi,natz3
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz2
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@3230:         mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@3230
                mov     [ebx],edx
                dec     loc_lb
                jz      @@3240
@@3250:         add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@3260:         mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@3260
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@3250
@@3240:         lea     esi,natz2
                lea     edi,natz3
                mov     ecx,pm
@@3270:         cmp     dword ptr[esi+ecx*4],0
                jne     @@3280
                dec     ecx
                jnz     @@3270
                inc     ecx
@@3280:         mov     [edi],ecx
@@3290:         mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@3290
                jmp     @@3130
@@3110:         mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                lea     esi,QT
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@4100
@@4110:         lea     edi,NT
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@4120
@@4100:         mov     la2,edx
                cmp     edx,2
                ja      @@4130
                je      @@4140
                mov     ebx,[esi+4]
                lea     esi,NT
                xor     edx,edx
@@4150:         mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@4150
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@4160
                dec     ecx
                jnz     @@4160
                inc     ecx
@@4160:         mov     [esi],ecx
                jmp     @@4120
@@4140:         mov     ebx,[esi+4]
                mov     eax,[esi+8]
                lea     esi,NT
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@4170:         sub     esi,4
                cmp     edx,reg_b
                je      @@4180
                div     reg_b
                jmp     @@4190
@@4180:         add     edx,eax
                mov     eax,4294967295
                jnc     @@4190
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@4200
                mov     ebx,[edi]
@@4200:         add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@4210
@@4190:         mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@4220
                mov     ebx,[edi]
@@4220:         shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@4230
                ja      @@4240
                test    eax,eax
                jz      @@4230
@@4240:         dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@4230
                ja      @@4240
                test    eax,eax
                jnz     @@4240
@@4230:         neg     eax
                adc     edx,0
                neg     edx
@@4210:         popf
                jnz     @@4170
                lea     edi,NT
                mov     ecx,la1
@@4250:         cmp     dword ptr [edi+ecx*4],0
                jne     @@4260
                dec     ecx
                jnz     @@4250
                inc     ecx
@@4260:         mov     [edi],ecx
                jmp     @@4120
@@4130:         mov     edx,la1
                lea     esi,natz1
                lea     ebx,NT
@@4270:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@4270
@@4280:         lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                lea     edi,QT
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@4290
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@4300
@@4290:         lea     ebx,NT
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@4310
@@4300:         mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@4320:         sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@4330
                div     reg_a
                mov     reg_q,eax
                jmp     @@4340
@@4330:         add     edx,eax
                mov     reg_q,4294967295
                jc      @@4350
@@4340:         mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@4350
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@4350
                dec     reg_q
@@4350:         mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@4360:         mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@4360
                cmp     [esi],ebx
                je      @@4370
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@4380:         mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@4380
@@4370:         lea     ebx,NT
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@4320
@@4310:         mov     ecx,la1
@@4390:         cmp     dword ptr[ebx+ecx*4],0
                jne     @@4400
                dec     ecx
                jnz     @@4390
                inc     ecx
@@4400:         mov     [ebx],ecx
@@4120:         lea     esi,QT
                lea     edi,UT
                mov     ecx,[edi]
                mov     [esi],ecx
                add     edi,4
                add     esi,4
                xor     ebx,ebx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                neg     ecx
@@4500:         mov     eax,[edi+ecx*4]
                mul     cn
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jne     @@4500
                cmp     ebx,0
                je      @@4510
                lea     esi,QT
                inc     dword ptr [esi]
                mov     ecx,[esi]
                mov     [esi+ecx*4],ebx
@@4510:         lea     esi,NT
                mov     edx,[esi]
                add     esi,4
                lea     edi,QT
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@5100
                xchg    edi,esi
                xchg    edx,ecx
@@5100:         lea     ebx,natz1
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@5110
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@5120:         mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@5120
                jnc     @@5110
                test    edx,edx
                jz      @@5130
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@5140:         mov     eax,[esi+edx*4]
                inc     eax
                jne     @@5150
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@5140
@@5130:         mov     dword ptr [ebx],1
                lea     ebx,natz1
                inc     dword ptr [ebx]
                jmp     @@5160
@@5110:         test    edx,edx
                jz      @@5170
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@5180:         mov     eax,[esi+edx*4]
@@5150:         mov     [ebx+edx*4],eax
                inc     edx
                jne     @@5180
@@5170:         mov     [ebx],ecx
@@5160:         lea     edi,natz1
                mov     ebx,[ebp+12]
                lea     esi,UT
                mov     ecx,[edi]
                xor     edx,edx
@@6010:         mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@6010
                mov     ecx,[edi]
                cmp     dword ptr[esi+ecx*4],0
                jne     @@6020
                dec     ecx
                jnz     @@6020
                inc     ecx
@@6020:         mov     [esi],ecx
                mov     esi,[ebp+16]
                lea     edi,UT
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@2800
                ja      @@9900
@@7001:         mov     eax,[esi+ecx*4]
                cmp     eax,[edi+ecx*4]
                ja      @@2800
                jb      @@9900
                dec     ecx
                jnz     @@7001
@@9900:
end;

procedure npow_1 (a: nat; n: longword; var b: nat); assembler;
asm
                mov     esi,[ebp+16]
                mov     edx,[ebp+12]
                dec     edx
                jns     @@100
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],1
                jmp     @@110
@@100:          mov     edi,[ebp+8]
                lea     ebx,natz3
                mov     ecx,[edi]
                mov     [esi],ecx
                mov     [ebx],ecx
@@120:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@120
                mov     kz,edx
@@130:          test    kz,1
                jz      @@140
                mov     esi,[ebp+16]
                mov     ecx,[esi]
                lea     edi,natz3
                mov     edx,[edi]
                lea     ebx,natz1
                cmp     ecx,edx
                jae     @@150
                xchg    edi,esi
                xchg    edx,ecx
@@150:          mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@160:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@160
                mov     [ebx],edx
                dec     loc_lb
                jz      @@170
@@180:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@190:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@190
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@180
@@170:          lea     esi,natz1
                mov     edi,[ebp+16]
                mov     ecx,pm
@@200:          cmp     dword ptr[esi+ecx*4],0
                jne     @@210
                dec     ecx
                jnz     @@200
                inc     ecx
@@210:          mov     [edi],ecx
@@220:          mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@220
@@140:          shr     kz,1
                jz      @@110
                lea     esi,natz3
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz2
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@230:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@230
                mov     [ebx],edx
                dec     loc_lb
                jz      @@240
@@250:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@260:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@260
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@250
@@240:          lea     esi,natz2
                lea     edi,natz3
                mov     ecx,pm
@@270:          cmp     dword ptr[esi+ecx*4],0
                jne     @@280
                dec     ecx
                jnz     @@270
                inc     ecx
@@280:          mov     [edi],ecx
@@290:          mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@290
                jmp     @@130
@@110:
end;

procedure nmodexp (a: nat; b: nat; c: nat; var d: nat); assembler;
asm
                mov     edi,[ebp+8]
                mov     edx,[ebp+12]
                mov     ebx,[ebp+20]
                lea     esi,natz1
                mov     ecx,[edi]
                mov     [esi],ecx
                test    dword ptr [edx+4],1
                jz      @@100
                mov     [ebx],ecx
@@101:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@101
                jmp     @@102
@@100:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@100
                mov     dword ptr [ebx],1
                mov     dword ptr [ebx+4],1
@@102:          mov     ecx,[edx]
                mov     eax,[edx+ecx*4]
                cmp     eax,0
                je      @@103
                bsr     eax,eax
                inc     eax
@@103:          dec     ecx
                shl     ecx,5
                add     eax,ecx
                dec     eax
                jnz     @@104
                mov     ecx,[edi]
                mov     nc,1
                jmp     @@105
@@104:          mov     nc,eax
                mov     kv,0
@@106:          inc     kv
                lea     esi,natz1
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz2
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@107:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@107
                mov     [ebx],edx
                dec     loc_lb
                jz      @@108
@@109:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@110:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@110
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@109
@@108:          lea     edi,natz2
                mov     ecx,pm
@@111:          cmp     dword ptr[edi+ecx*4],0
                jne     @@112
                dec     ecx
                jnz     @@111
                inc     ecx
@@112:          mov     [edi],ecx
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@113
                lea     esi,natz1
                mov     [esi],ecx
@@114:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@114
                jmp     @@115
@@113:          mov     la2,edx
                cmp     edx,2
                ja      @@116
                je      @@117
                mov     ebx,[esi+4]
                xor     edx,edx
@@118:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@118
                lea     ebx,natz1
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@115
@@117:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@119:          cmp     edx,reg_b
                je      @@120
                div     reg_b
                jmp     @@121
@@120:          add     edx,eax
                mov     eax,4294967295
                jnc     @@121
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@122
                mov     ebx,[edi]
@@122:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@123
@@121:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@124
                mov     ebx,[edi]
@@124:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@125
                ja      @@126
                test    eax,eax
                jz      @@125
@@126:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@125
                ja      @@126
                test    eax,eax
                jnz     @@126
@@125:          neg     eax
                adc     edx,0
                neg     edx
@@123:          popf
                jnz     @@119
                shrd    eax,edx,cl
                shr     edx,cl
                lea     edi,natz1
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@127
                mov     dword ptr [edi],1
                jmp     @@115
@@127:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@115
@@116:          mov     edx,la1
                lea     esi,natz1
                mov     [esi],ecx
@@128:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@128
@@129:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@130
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@131
@@130:          test    edx,edx
                jz      @@132
@@131:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@133:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@134
                div     reg_a
                mov     reg_q,eax
                jmp     @@135
@@134:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@136
@@135:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@136
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@136
                dec     reg_q
@@136:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@137:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@137
                cmp     [esi],ebx
                je      @@138
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@139:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@139
@@138:          dec     reg_c
                jne     @@133
@@132:          lea     esi,natz1
                mov     ecx,la2
@@140:          cmp     dword ptr[esi+ecx*4],0
                jne     @@141
                dec     ecx
                jnz     @@140
                inc     ecx
@@141:          mov     [esi],ecx
@@115:          mov     edi,[ebp+12]
                mov     edx,kv
                mov     ecx,edx
                and     ecx,31
                shr     edx,5
                inc     edx
                mov     eax,[edi+edx*4]
                bt      eax,ecx
                jnc     @@142
                mov     esi,[ebp+20]
                mov     ecx,[esi]
                lea     edi,natz1
                mov     edx,[edi]
                lea     ebx,natz2
                cmp     ecx,edx
                jae     @@143
                xchg    edi,esi
                xchg    edx,ecx
@@143:          mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@144:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@144
                mov     [ebx],edx
                dec     loc_lb
                jz      @@145
@@146:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@147:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@147
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@146
@@145:          lea     edi,natz2
                mov     ecx,pm
@@148:          cmp     dword ptr[edi+ecx*4],0
                jne     @@149
                dec     ecx
                jnz     @@148
                inc     ecx
@@149:          mov     [edi],ecx
@@105:          mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@150
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@151:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@151
                jmp     @@142
@@150:          mov     la2,edx
                cmp     edx,2
                ja      @@152
                je      @@153
                mov     ebx,[esi+4]
                xor     edx,edx
@@154:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@154
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@142
@@153:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@155:          cmp     edx,reg_b
                je      @@156
                div     reg_b
                jmp     @@157
@@156:          add     edx,eax
                mov     eax,4294967295
                jnc     @@157
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@158
                mov     ebx,[edi]
@@158:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@159
@@157:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@160
                mov     ebx,[edi]
@@160:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@161
                ja      @@162
                test    eax,eax
                jz      @@161
@@162:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@161
                ja      @@162
                test    eax,eax
                jnz     @@162
@@161:          neg     eax
                adc     edx,0
                neg     edx
@@159:          popf
                jnz     @@155
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@163
                mov     dword ptr [edi],1
                jmp     @@142
@@163:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@142
@@152:          mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@164:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@164
@@165:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@166
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@167
@@166:          test    edx,edx
                jz      @@168
@@167:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@169:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@170
                div     reg_a
                mov     reg_q,eax
                jmp     @@171
@@170:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@172
@@171:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@172
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@172
                dec     reg_q
@@172:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@173:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@173
                cmp     [esi],ebx
                je      @@174
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@175:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@175
@@174:          dec     reg_c
                jne     @@169
@@168:          mov     esi,[ebp+20]
                mov     ecx,la2
@@176:          cmp     dword ptr[esi+ecx*4],0
                jne     @@177
                dec     ecx
                jnz     @@176
                inc     ecx
@@177:          mov     [esi],ecx
@@142:          dec     nc
                jnz     @@106
@@178:
end;

procedure npw2 (var a: nat; n: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     eax,[ebp+12]
                mov     ecx,eax
                shr     eax,5
                inc     eax
                mov     [edi],eax
                and     ecx,31
                mov     ebx,1
                shl     ebx,cl
                mov     [edi+eax*4],ebx
                dec     eax
                jz      @@2
@@1:            mov     dword ptr [edi+eax*4],0
                dec     eax
                jnz     @@1
@@2:
end;

procedure naddmod (a,b,c: nat; var d: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@100
                xchg    edi,esi
                xchg    edx,ecx
@@100:          lea     ebx,natz1
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@110
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@120:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@120
                jnc     @@110
                test    edx,edx
                jz      @@130
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@140:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@150
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@140
@@130:          mov     dword ptr [ebx],1
                lea     ebx,natz1
                inc     dword ptr [ebx]
                jmp     @@160
@@110:          test    edx,edx
                jz      @@170
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@180:          mov     eax,[esi+edx*4]
@@150:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@180
@@170:          mov     [ebx],ecx
@@160:          lea     edi,natz1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@200
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@210:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@210
                jmp     @@220
@@200:          mov     la2,edx
                cmp     edx,2
                ja      @@230
                je      @@240
                mov     ebx,[esi+4]
                xor     edx,edx
@@250:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@250
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@220
@@240:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@260:          cmp     edx,reg_b
                je      @@270
                div     reg_b
                jmp     @@280
@@270:          add     edx,eax
                mov     eax,4294967295
                jnc     @@280
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@290
                mov     ebx,[edi]
@@290:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@300
@@280:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@310
                mov     ebx,[edi]
@@310:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@320
                ja      @@330
                test    eax,eax
                jz      @@320
@@330:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@320
                ja      @@330
                test    eax,eax
                jnz     @@330
@@320:          neg     eax
                adc     edx,0
                neg     edx
@@300:          popf
                jnz     @@260
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@340
                mov     dword ptr [edi],1
                jmp     @@220
@@340:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@220
@@230:          mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@350:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@350
@@360:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@370
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@380
@@370:          test    edx,edx
                jz      @@390
@@380:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@400:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@410
                div     reg_a
                mov     reg_q,eax
                jmp     @@420
@@410:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@430
@@420:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@430
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@430
                dec     reg_q
@@430:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@440:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@440
                cmp     [esi],ebx
                je      @@450
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@460:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@460
@@450:          dec     reg_c
                jne     @@400
@@390:          mov     esi,[ebp+20]
                mov     ecx,la2
@@470:          cmp     dword ptr[esi+ecx*4],0
                jne     @@480
                dec     ecx
                jnz     @@470
                inc     ecx
@@480:          mov     [esi],ecx
@@220:
end;

procedure nsubmod (a,b,c: nat; var d: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+12]
                mov     ecx,[edi]
                add     edi,4
                lea     ebx,natz1
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@100
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@101:          mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@101
                jnb     @@100
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@102:          mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@103
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@102
                jmp     @@104
@@100:          test    edx,edx
                jz      @@104
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@105:          mov     eax,[esi+edx*4]
@@103:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@105
@@104:          lea     ebx,natz1
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jne     @@106
                dec     ecx
                jz      @@106
@@107:          cmp     dword ptr[ebx+ecx*4],0
                jne     @@108
                dec     ecx
                jnz     @@107
                inc     ecx
@@108:          mov     [ebx],ecx
@@106:          lea     edi,natz1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@109
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@110:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@110
                jmp     @@111
@@109:          mov     la2,edx
                cmp     edx,2
                ja      @@112
                je      @@113
                mov     ebx,[esi+4]
                xor     edx,edx
@@114:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@114
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@111
@@113:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@115:          cmp     edx,reg_b
                je      @@116
                div     reg_b
                jmp     @@117
@@116:          add     edx,eax
                mov     eax,4294967295
                jnc     @@117
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@118
                mov     ebx,[edi]
@@118:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@119
@@117:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@120
                mov     ebx,[edi]
@@120:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@121
                ja      @@122
                test    eax,eax
                jz      @@121
@@122:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@121
                ja      @@122
                test    eax,eax
                jnz     @@122
@@121:          neg     eax
                adc     edx,0
                neg     edx
@@119:          popf
                jnz     @@115
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@123
                mov     dword ptr [edi],1
                jmp     @@111
@@123:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@111
@@112:          mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@124:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@124
@@125:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@126
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@127
@@126:          test    edx,edx
                jz      @@128
@@127:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@129:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@130
                div     reg_a
                mov     reg_q,eax
                jmp     @@131
@@130:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@132
@@131:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@132
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@132
                dec     reg_q
@@132:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@133:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@133
                cmp     [esi],ebx
                je      @@134
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@135:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@135
@@134:          dec     reg_c
                jne     @@129
@@128:          mov     esi,[ebp+20]
                mov     ecx,la2
@@136:          cmp     dword ptr[esi+ecx*4],0
                jne     @@137
                dec     ecx
                jnz     @@136
                inc     ecx
@@137:          mov     [esi],ecx
@@111:
end;

procedure nmulmod (a,b,c: nat; var d: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                mov     edx,[edi]
                lea     ebx,natz1
                cmp     ecx,edx
                jae     @@100
                xchg    edi,esi
                xchg    edx,ecx
@@100:          mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@110:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@110
                mov     [ebx],edx
                dec     loc_lb
                jz      @@120
@@130:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@140:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@140
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@130
@@120:          lea     esi,natz1
                mov     ecx,pm
@@150:          cmp     dword ptr[esi+ecx*4],0
                jne     @@160
                dec     ecx
                jnz     @@150
                inc     ecx
@@160:          mov     [esi],ecx
                lea     edi,natz1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@200
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@210:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@210
                jmp     @@220
@@200:          mov     la2,edx
                cmp     edx,2
                ja      @@230
                je      @@240
                mov     ebx,[esi+4]
                xor     edx,edx
@@250:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@250
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@220
@@240:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@260:          cmp     edx,reg_b
                je      @@270
                div     reg_b
                jmp     @@280
@@270:          add     edx,eax
                mov     eax,4294967295
                jnc     @@280
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@290
                mov     ebx,[edi]
@@290:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@300
@@280:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@310
                mov     ebx,[edi]
@@310:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@320
                ja      @@330
                test    eax,eax
                jz      @@320
@@330:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@320
                ja      @@330
                test    eax,eax
                jnz     @@330
@@320:          neg     eax
                adc     edx,0
                neg     edx
@@300:          popf
                jnz     @@260
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@340
                mov     dword ptr [edi],1
                jmp     @@220
@@340:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@220
@@230:          mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@350:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@350
@@360:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@370
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@380
@@370:          test    edx,edx
                jz      @@390
@@380:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@400:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@410
                div     reg_a
                mov     reg_q,eax
                jmp     @@420
@@410:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@430
@@420:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@430
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@430
                dec     reg_q
@@430:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@440:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@440
                cmp     [esi],ebx
                je      @@450
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@460:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@460
@@450:          dec     reg_c
                jne     @@400
@@390:          mov     esi,[ebp+20]
                mov     ecx,la2
@@470:          cmp     dword ptr[esi+ecx*4],0
                jne     @@480
                dec     ecx
                jnz     @@470
                inc     ecx
@@480:          mov     [esi],ecx
@@220:
end;

procedure nsqrmod (a,b: nat; var c: nat); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz1
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@100:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@100
                mov     [ebx],edx
                dec     loc_lb
                jz      @@110
@@120:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@130:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@130
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@120
@@110:          lea     edi,natz1
                mov     ecx,pm
@@140:          cmp     dword ptr[edi+ecx*4],0
                jne     @@150
                dec     ecx
                jnz     @@140
                inc     ecx
@@150:          mov     [edi],ecx
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@200
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@210:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@210
                jmp     @@220
@@200:          mov     la2,edx
                cmp     edx,2
                ja      @@230
                je      @@240
                mov     ebx,[esi+4]
                xor     edx,edx
@@250:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@250
                mov     ebx,[ebp+16]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@220
@@240:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@260:          cmp     edx,reg_b
                je      @@270
                div     reg_b
                jmp     @@280
@@270:          add     edx,eax
                mov     eax,4294967295
                jnc     @@280
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@290
                mov     ebx,[edi]
@@290:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@300
@@280:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@310
                mov     ebx,[edi]
@@310:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@320
                ja      @@330
                test    eax,eax
                jz      @@320
@@330:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@320
                ja      @@330
                test    eax,eax
                jnz     @@330
@@320:          neg     eax
                adc     edx,0
                neg     edx
@@300:          popf
                jnz     @@260
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+16]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@340
                mov     dword ptr [edi],1
                jmp     @@220
@@340:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@220
@@230:          mov     edx,la1
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@350:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@350
@@360:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@370
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@380
@@370:          test    edx,edx
                jz      @@390
@@380:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@400:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@410
                div     reg_a
                mov     reg_q,eax
                jmp     @@420
@@410:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@430
@@420:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@430
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@430
                dec     reg_q
@@430:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@440:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@440
                cmp     [esi],ebx
                je      @@450
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@460:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@460
@@450:          dec     reg_c
                jne     @@400
@@390:          mov     esi,[ebp+16]
                mov     ecx,la2
@@470:          cmp     dword ptr[esi+ecx*4],0
                jne     @@480
                dec     ecx
                jnz     @@470
                inc     ecx
@@480:          mov     [esi],ecx
@@220:
end;

procedure ngcd (a: nat; b: nat; var c: nat); assembler;
asm
                lea     eax,natz3
                mov     ofs1,eax
                lea     ebx,natz1
                mov     ofs2,ebx
                lea     edx,natz2
                mov     ofs3,edx
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                cmp     dword ptr [esi],1
                jne     @@100
                cmp     dword ptr [esi+4],0
                jne     @@100
                mov     esi,[ebp+16]
                mov     ecx,[edi]
                mov     [esi],ecx
@@101:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@101
                jmp     @@102
@@100:          mov     ecx,[edi]
                mov     [ebx],ecx
@@103:          mov     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@103
                mov     ecx,[esi]
                mov     [edx],ecx
@@104:          mov     eax,[esi+ecx*4]
                mov     [edx+ecx*4],eax
                dec     ecx
                jnz     @@104
@@105:          mov     eax,ofs1
                mov     ebx,ofs2
                mov     ecx,ofs3
                mov     ofs1,ebx
                mov     ofs2,ecx
                mov     ofs3,eax
                mov     edi,ofs1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,ofs2
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@106
                mov     esi,ofs3
                mov     [esi],ecx
@@107:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@107
                jmp     @@108
@@106:          mov     la2,edx
                cmp     edx,2
                ja      @@109
                je      @@110
                mov     ebx,[esi+4]
                xor     edx,edx
@@111:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@111
                mov     ebx,ofs3
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@108
@@110:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@112:          cmp     edx,reg_b
                je      @@113
                div     reg_b
                jmp     @@114
@@113:          add     edx,eax
                mov     eax,4294967295
                jnc     @@114
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@115
                mov     ebx,[edi]
@@115:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@116
@@114:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@117
                mov     ebx,[edi]
@@117:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@118
                ja      @@119
                test    eax,eax
                jz      @@118
@@119:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@118
                ja      @@119
                test    eax,eax
                jnz     @@119
@@118:          neg     eax
                adc     edx,0
                neg     edx
@@116:          popf
                jnz     @@112
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,ofs3
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@120
                mov     dword ptr [edi],1
                jmp     @@108
@@120:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@108
@@109:          mov     edx,la1
                mov     esi,ofs3
                mov     [esi],ecx
@@121:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@121
@@122:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,ofs2
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@123
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@124
@@123:          test    edx,edx
                jz      @@125
@@124:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@126:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@127
                div     reg_a
                mov     reg_q,eax
                jmp     @@128
@@127:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@129
@@128:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@129
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@129
                dec     reg_q
@@129:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@130:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@130
                cmp     [esi],ebx
                je      @@131
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@132:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@132
@@131:          dec     reg_c
                jne     @@126
@@125:          mov     esi,ofs3
                mov     ecx,la2
@@133:          cmp     dword ptr[esi+ecx*4],0
                jne     @@134
                dec     ecx
                jnz     @@133
                inc     ecx
@@134:          mov     [esi],ecx
@@108:          mov     esi,ofs3
                cmp     dword ptr [esi],1
                jne     @@105
                cmp     dword ptr [esi+4],0
                jne     @@105
                mov     esi,[ebp+16]
                mov     edi,ofs2
                mov     ecx,[edi]
                mov     [esi],ecx
@@135:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@135
@@102:
end;

function ngcd0 (a: nat; n: longword): longword;
begin
if n=0 then ngcd0:=0
else begin
dl:=nmod0(a,n);
if dl=0 then ngcd0:=n
else begin
cl:=n;
repeat
el:=cl;
cl:=dl;
dl:=el mod dl;
until dl=0;
ngcd0:=cl
end
end
end;

procedure nlcm (a,b: nat; var c: nat);
begin
ngcd(a,b,c);
nmov(a,ut);
if (c[0]>1) or (c[1]>1) then nquos(ut,c);
nmul(ut,b,c)
end;

function njacobi (a,b: nat): longint; assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@1000
                lea     esi,natz2
                mov     [esi],ecx
@@1010:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1010
                jmp     @@1020
@@1000:         mov     la2,edx
                cmp     edx,2
                ja      @@1030
                je      @@1040
                mov     ebx,[esi+4]
                xor     edx,edx
@@1050:         mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1050
                lea     ebx,natz2
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@1020
@@1040:         mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@1060:         cmp     edx,reg_b
                je      @@1070
                div     reg_b
                jmp     @@1080
@@1070:         add     edx,eax
                mov     eax,4294967295
                jnc     @@1080
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@1090
                mov     ebx,[edi]
@@1090:         add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@1100
@@1080:         mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@1110
                mov     ebx,[edi]
@@1110:         shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@1120
                ja      @@1130
                test    eax,eax
                jz      @@1120
@@1130:         dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@1120
                ja      @@1130
                test    eax,eax
                jnz     @@1130
@@1120:         neg     eax
                adc     edx,0
                neg     edx
@@1100:         popf
                jnz     @@1060
                shrd    eax,edx,cl
                shr     edx,cl
                lea     edi,natz2
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@1140
                mov     dword ptr [edi],1
                jmp     @@1020
@@1140:         mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@1020
@@1030:         mov     edx,la1
                lea     esi,natz2
                mov     [esi],ecx
@@1150:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1150
@@1160:         lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@1170
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@1180
@@1170:         test    edx,edx
                jz      @@1190
@@1180:         mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@1200:         sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@1210
                div     reg_a
                mov     reg_q,eax
                jmp     @@1220
@@1210:         add     edx,eax
                mov     reg_q,4294967295
                jc      @@1230
@@1220:         mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@1230
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@1230
                dec     reg_q
@@1230:         mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@1240:         mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@1240
                cmp     [esi],ebx
                je      @@1250
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@1260:         mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@1260
@@1250:         dec     reg_c
                jne     @@1200
@@1190:         lea     esi,natz2
                mov     ecx,la2
@@1270:         cmp     dword ptr[esi+ecx*4],0
                jne     @@1280
                dec     ecx
                jnz     @@1270
                inc     ecx
@@1280:         mov     [esi],ecx
@@1020:         lea     esi,natz2
                cmp     dword ptr [esi],1
                jne     @@1290
                cmp     dword ptr [esi+4],0
                je      @@1300
@@1290:         mov     edi,[ebp+12]
                lea     esi,natz3
                mov     ecx,[edi]
                mov     [esi],ecx
@@1310:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1310
                mov     kv,1
@@1320:         lea     esi,natz2
                mov     ecx,[esi]
                xor     edx,edx
                lea     esi,[esi+ecx*4+4]
                neg     ecx
                mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@1330
                inc     ecx
                jz      @@1340
                mov     edx,32
@@1350:         mov     eax,[esi+ecx*4]
                cmp     eax,0
                jne     @@1330
                add     edx,32
                inc     ecx
                jnz     @@1350
                xor     edx,edx
                jmp     @@1340
@@1330:         bsf     ecx,eax
                add     edx,ecx
@@1340:         mov     reg_a,edx
                lea     edi,natz2
                mov     esi,edi
                mov     ecx,edx
                mov     edx,[edi]
                cmp     ecx,0
                je      @@1360
@@1370:         cmp     ecx,32
                jb      @@1380
                mov     eax,ecx
                shr     eax,5
                cmp     eax,edx
                jb      @@1390
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@1360
@@1390:         and     ecx,31
                jnz     @@1400
                mov     ecx,edx
                sub     ecx,eax
                mov     [edi],ecx
                lea     esi,natz1
@@1410:         mov     eax,[edi+edx*4]
                mov     [esi+ecx*4],eax
                dec     edx
                dec     ecx
                jnz     @@1410
                mov     ecx,[edi]
@@1420:         mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@1420
                jmp     @@1360
@@1400:         sub     edx,eax
                lea     edi,[edi+eax*4]
@@1380:         mov     [esi],edx
                lea     edi,[edi+edx*4+4]
                lea     esi,[esi+edx*4]
                neg     edx
                mov     ebx,[edi+edx*4]
                inc     edx
                jnz     @@1430
                shr     ebx,cl
                mov     [esi],ebx
                jmp     @@1360
@@1430:         mov     eax,[edi+edx*4]
                shrd    ebx,eax,cl
                mov     [esi+edx*4],ebx
                mov     ebx,eax
                inc     edx
                jnz     @@1430
                shr     ebx,cl
                mov     [esi],ebx
                lea     edi,natz2
@@1440:         mov     ecx,[edi]
                cmp     dword ptr [edi+ecx*4],0
                jne     @@1360
                dec     ecx
                jz      @@1360
@@1450:         cmp     dword ptr [edi+ecx*4],0
                jne     @@1460
                dec     ecx
                jnz     @@1450
                inc     ecx
@@1460:         mov     [edi],ecx
@@1360:         mov     eax,reg_a
                test    eax,1
                jz      @@1470
                lea     edi,natz3
                mov     eax,[edi+4]
                and     eax,7
                cmp     al,3
                je      @@1480
                cmp     al,5
                jne     @@1470
@@1480:         neg     kv
@@1470:         lea     edi,natz2
                cmp     dword ptr [edi],1
                jne     @@1490
                cmp     dword ptr [edi+4],1
                jne     @@1490
                cmp     kv,1
                jne     @@1500
                mov     eax,1
                jmp     @@1510
@@1500:         mov     eax,4294967295
                jmp     @@1510
@@1490:         lea     edi,natz3
                mov     ecx,[edi]
                mov     la1,ecx
                lea     esi,natz2
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@1520
                lea     esi,natz4
                mov     [esi],ecx
@@1530:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1530
                jmp     @@1540
@@1520:         mov     la2,edx
                cmp     edx,2
                ja      @@1550
                je      @@1560
                mov     ebx,[esi+4]
                xor     edx,edx
@@1570:         mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1570
                lea     ebx,natz4
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@1540
@@1560:         mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@1580:         cmp     edx,reg_b
                je      @@1590
                div     reg_b
                jmp     @@1600
@@1590:         add     edx,eax
                mov     eax,4294967295
                jnc     @@1600
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@1610
                mov     ebx,[edi]
@@1610:         add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@1620
@@1600:         mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@1630
                mov     ebx,[edi]
@@1630:         shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@1640
                ja      @@1650
                test    eax,eax
                jz      @@1640
@@1650:         dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@1640
                ja      @@1650
                test    eax,eax
                jnz     @@1650
@@1640:         neg     eax
                adc     edx,0
                neg     edx
@@1620:         popf
                jnz     @@1580
                shrd    eax,edx,cl
                shr     edx,cl
                lea     edi,natz4
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@1660
                mov     dword ptr [edi],1
                jmp     @@1540
@@1660:         mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@1540
@@1550:         mov     edx,la1
                lea     esi,natz4
                mov     [esi],ecx
@@1670:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1670
@@1680:         lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                lea     edi,natz2
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@1690
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@1700
@@1690:         test    edx,edx
                jz      @@1710
@@1700:         mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@1720:         sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@1730
                div     reg_a
                mov     reg_q,eax
                jmp     @@1740
@@1730:         add     edx,eax
                mov     reg_q,4294967295
                jc      @@1750
@@1740:         mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@1750
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@1750
                dec     reg_q
@@1750:         mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@1760:         mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@1760
                cmp     [esi],ebx
                je      @@1770
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@1780:         mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@1780
@@1770:         dec     reg_c
                jne     @@1720
@@1710:         lea     esi,natz4
                mov     ecx,la2
@@1790:         cmp     dword ptr[esi+ecx*4],0
                jne     @@1800
                dec     ecx
                jnz     @@1790
                inc     ecx
@@1800:         mov     [esi],ecx
@@1540:         lea     edi,natz4
                cmp     dword ptr [edi],1
                jne     @@1810
                cmp     dword ptr [edi+4],0
                je      @@1300
@@1810:         lea     edi,natz3
                mov     eax,[edi+4]
                and     eax,3
                cmp     al,3
                jne     @@1820
                lea     edi,natz2
                mov     eax,[edi+4]
                and     eax,3
                cmp     al,3
                jne     @@1820
                neg     kv
@@1820:         lea     edi,natz2
                lea     esi,natz3
                lea     ebx,natz4
                mov     ecx,[edi]
                mov     [esi],ecx
@@1830:         mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1830
                mov     ecx,[ebx]
                mov     [edi],ecx
@@1840:         mov     eax,[ebx+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@1840
                jmp     @@1320
@@1300:         xor     eax,eax
@@1510:
end;

procedure fibn (n: longword; var a: nat); assembler;
asm
                mov     esi,[ebp+12]
                mov     dword ptr [esi],1
                mov     eax,[ebp+8]
                cmp     eax,0
                jne     @@100
                mov     dword ptr [esi+4],0
                jmp     @@101
@@100:          mov     dword ptr [esi+4],1
                cmp     eax,3
                jb      @@101
                lea     edi,natz1
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                bsr     eax,eax
                dec     eax
                mov     kv,eax
@@102:          mov     esi,[ebp+12]
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz2
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@103:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@103
                mov     [ebx],edx
                dec     loc_lb
                jz      @@104
@@105:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@106:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@106
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@105
@@104:          lea     esi,natz2
                mov     ecx,pm
@@107:          cmp     dword ptr[esi+ecx*4],0
                jne     @@108
                dec     ecx
                jnz     @@107
                inc     ecx
@@108:          mov     [esi],ecx
                lea     esi,natz1
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,ut
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@109:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@109
                mov     [ebx],edx
                dec     loc_lb
                jz      @@110
@@111:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@112:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@112
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@111
@@110:          lea     esi,ut
                mov     ecx,pm
@@113:          cmp     dword ptr[esi+ecx*4],0
                jne     @@114
                dec     ecx
                jnz     @@113
                inc     ecx
@@114:          mov     [esi],ecx
                lea     esi,natz2
                mov     edx,[esi]
                add     esi,4
                lea     edi,ut
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@115
                xchg    edi,esi
                xchg    edx,ecx
@@115:          lea     ebx,natz3
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@116
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@117:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@117
                jnc     @@116
                test    edx,edx
                jz      @@118
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@119:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@120
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@119
@@118:          mov     dword ptr [ebx],1
                lea     ebx,natz3
                inc     dword ptr [ebx]
                jmp     @@121
@@116:          test    edx,edx
                jz      @@122
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@123:          mov     eax,[esi+edx*4]
@@120:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@123
@@122:          mov     [ebx],ecx
@@121:          lea     esi,natz1
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                mov     edx,[edi]
                lea     ebx,ut
                cmp     ecx,edx
                jae     @@124
                xchg    edi,esi
                xchg    edx,ecx
@@124:          mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@125:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@125
                mov     [ebx],edx
                dec     loc_lb
                jz      @@126
@@127:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@128:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@128
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@127
@@126:          lea     esi,ut
                mov     ecx,pm
@@129:          cmp     dword ptr[esi+ecx*4],0
                jne     @@130
                dec     ecx
                jnz     @@129
                inc     ecx
@@130:          mov     [esi],ecx
                lea     edi,ut
                mov     ecx,[edi]
                xor     ebx,ebx
                clc
@@131:          inc     ebx
                rcl     dword ptr [edi+ebx*4],1
                dec     ecx
                jnz     @@131
                jnc     @@132
                inc     ebx
                mov     dword ptr [edi+ebx*4],1
                inc     dword ptr [edi]
@@132:          lea     esi,natz2
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@133
                xchg    edi,esi
                xchg    edx,ecx
@@133:          lea     ebx,natz4
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@134
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@135:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@135
                jnc     @@134
                test    edx,edx
                jz      @@136
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@137:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@138
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@137
@@136:          mov     dword ptr [ebx],1
                lea     ebx,natz4
                inc     dword ptr [ebx]
                jmp     @@139
@@134:          test    edx,edx
                jz      @@140
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@141:          mov     eax,[esi+edx*4]
@@138:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@141
@@140:          mov     [ebx],ecx
@@139:          mov     eax,[ebp+8]
                mov     ebx,kv
                bt      eax,ebx
                jnc     @@142
                lea     esi,natz3
                mov     edx,[esi]
                add     esi,4
                lea     edi,natz4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@143
                xchg    edi,esi
                xchg    edx,ecx
@@143:          mov     ebx,[ebp+12]
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@144
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@145:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@145
                jnc     @@144
                test    edx,edx
                jz      @@146
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@147:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@148
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@147
@@146:          mov     dword ptr [ebx],1
                mov     ebx,[ebp+12]
                inc     dword ptr [ebx]
                jmp     @@149
@@144:          test    edx,edx
                jz      @@150
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@151:          mov     eax,[esi+edx*4]
@@148:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@151
@@150:          mov     [ebx],ecx
@@149:          lea     edi,natz4
                lea     esi,natz1
                mov     ecx,[edi]
                mov     [esi],ecx
@@152:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@152
                jmp     @@153
@@142:          lea     edi,natz3
                lea     esi,natz1
                mov     ecx,[edi]
                mov     [esi],ecx
@@154:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@154
                lea     edi,natz4
                mov     esi,[ebp+12]
                mov     ecx,[edi]
                mov     [esi],ecx
@@155:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@155
@@153:          dec     kv
                jns     @@102
@@101:
end;

procedure cbin (n,k: longword; var a: nat); assembler;
asm
                mov     esi,[ebp+16]
                mov     dword ptr [esi],1
                mov     eax,[ebp+8]
                mov     ebx,[ebp+12]
                cmp     ebx,0
                je      @@10
                cmp     eax,ebx
                jne     @@11
@@10:           mov     dword ptr [esi+4],1
                jmp     @@12
@@11:           ja      @@13
                mov     dword ptr [esi+4],0
                jmp     @@12
@@13:           mov     [esi+4],eax
                cmp     ebx,1
                je      @@12
                mov     ecx,ebx
                inc     ecx
                cmp     eax,ecx
                je      @@12
                mov     edi,eax
                mov     ecx,ebx
                shl     ecx,1
                cmp     eax,ecx
                jae     @@14
                sub     eax,ebx
                mov     ebx,eax
@@14:           dec     ebx
                mov     rc,ebx
                mov     j,1
@@15:           dec     edi
                mov     esi,[ebp+16]
                mov     ecx,[esi]
                xor     ebx,ebx
                lea     esi,[esi+ecx*4+4]
                neg     ecx
@@16:           mov     eax,[esi+ecx*4]
                mul     edi
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jnz     @@16
                cmp     ebx,0
                je      @@17
                mov     esi,[ebp+16]
                mov     eax,[esi]
                inc     eax
                cmp     eax,dimnat
                jbe     @@18
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@12
@@18:           mov     [esi],eax
                mov     [esi+eax*4],ebx
@@17:           inc     j
                mov     ebx,j
                mov     esi,[ebp+16]
                mov     ecx,[esi]
                xor     edx,edx
@@19:           mov     eax,[esi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@19
                mov     ecx,[esi]
                cmp     dword ptr[esi+ecx*4],0
                jne     @@20
                dec     dword ptr [esi]
                jnz     @@20
                inc     dword ptr [esi]
@@20:           dec     rc
                jnz     @@15
@@12:
end;

procedure factr (n: longword; var a: nat); assembler;
asm
                mov     esi,[ebp+12]
                mov     dword ptr [esi],1
                mov     edi,[ebp+8]
                cmp     edi,1
                ja      @@1
                mov     dword ptr [esi+4],1
                jmp     @@6
@@1:            mov     [esi+4],edi
                dec     edi
                cmp     edi,1
                je      @@6
                mov     ecx,[esi]
                xor     ebx,ebx
@@2:            lea     esi,[esi+ecx*4+4]
                neg     ecx
@@3:            mov     eax,[esi+ecx*4]
                mul     edi
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jnz     @@3
                mov     esi,[ebp+12]
                mov     ecx,[esi]
                cmp     ebx,0
                je      @@5
                inc     ecx
                cmp     ecx,dimnat
                jbe     @@4
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                jmp     @@6
@@4:            inc     dword ptr [esi]
                mov     [esi+ecx*4],ebx
                xor     ebx,ebx
@@5:            dec     edi
                jnz     @@2
@@6:
end;

{* signed integers *}

procedure strtoz (s: string; var a: zin);
begin
if s[1]='-' then begin
strton(copy(s,2,length(s)-1),a.v);
a.s:=-1
end
else begin
strton(s,a.v);
a.s:=1
end
end;

procedure ztostr (a: zin; var s: string);
begin
ntostr(a.v,s);
if a.s=-1 then s:='-'+s
end;

procedure zmov (a: zin; var b: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi]
                mov     [esi],ecx
@@1:            mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     eax,[edi+psign]
                mov     [esi+psign],eax
end;

procedure ltoz (l: longint; var a: zin); assembler;
asm
                mov     eax,[ebp+8]
                mov     edi,[ebp+12]
                mov     ebx,1
                mov     dword ptr [edi],ebx
                cmp     eax,2147483647
                jbe     @@1
                neg     eax
                neg     ebx
@@1:            mov     [edi+4],eax
                mov     dword ptr [edi+psign],ebx
end;

procedure i64toz (n: int64; var a: zin); assembler;
asm
                mov     eax,[ebp+8]
                mov     ebx,[ebp+12]
                mov     esi,[ebp+16]
                mov     ecx,1
                test    ebx,2147483648
                jz      @@1
                sub     eax,1
                sbb     ebx,0
                not     eax
                not     ebx
                neg     ecx
@@1:            mov     [esi+psign],ecx
                mov     [esi+4],eax
                mov     dword ptr [esi],1
                cmp     ebx,0
                je      @@2
                mov     [esi+8],ebx
                inc     dword ptr [esi]
@@2:
end;

function zcmp (a: zin; b: zin): longint; assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     eax,[edi+psign]
                cmp     eax,[esi+psign]
                jne     @@10
                mov     ebx,eax
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@11
                ja      @@12
@@13:           mov     eax,[edi+ecx*4]
                cmp     eax,[esi+ecx*4]
                jb      @@11
                ja      @@12
                dec     ecx
                jnz     @@13
                xor     eax,eax
                jmp     @@10
@@11:           mov     eax,4294967295
                jmp     @@14
@@12:           mov     eax,1
@@14:           cmp     ebx,1
                je      @@10
                neg     eax
@@10:
end;

function zequ (a: zin; b: zin): boolean; assembler;
asm
                xor     eax,eax
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi+psign]
                cmp     ecx,[esi+psign]
                jne     @@2
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jne     @@2
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jne     @@2
                dec     ecx
                jnz     @@1
                inc     eax
@@2:
end;

function zdif (a: zin; b: zin): boolean; assembler;
asm
                mov     eax,1
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi+psign]
                cmp     ecx,[esi+psign]
                jne     @@2
                mov     ecx,[edi]
                cmp     ecx,[esi]
                jne     @@2
@@1:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jne     @@2
                dec     ecx
                jnz     @@1
                dec     eax
@@2:
end;

function zlt (a: zin; b: zin): boolean; assembler;
asm
                mov     eax,1
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     edx,[edi+psign]
                cmp     edx,[esi+psign]
                je      @@1
                inc     edx
                jz      @@2
                jmp     @@3
@@1:            mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@4
                ja      @@5
@@6:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@4
                ja      @@5
                dec     ecx
                jnz     @@6
                jmp     @@3
@@4:            inc     edx
                jz      @@3
                jmp     @@2
@@5:            inc     edx
                jz      @@2
@@3:            dec     eax
@@2:
end;

function zlte (a: zin; b: zin): boolean; assembler;
asm
                mov     eax,1
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     edx,[edi+psign]
                cmp     edx,[esi+psign]
                je      @@1
                inc     edx
                jz      @@2
                jmp     @@3
@@1:            mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@4
                ja      @@5
@@6:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@4
                ja      @@5
                dec     ecx
                jnz     @@6
                jmp     @@2
@@4:            inc     edx
                jz      @@3
                jmp     @@2
@@5:            inc     edx
                jz      @@2
@@3:            dec     eax
@@2:
end;

function zgt (a: zin; b: zin): boolean; assembler;
asm
                xor     eax,eax
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     edx,[edi+psign]
                cmp     edx,[esi+psign]
                je      @@1
                inc     edx
                jz      @@2
                jmp     @@3
@@1:            mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@4
                ja      @@5
@@6:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@4
                ja      @@5
                dec     ecx
                jnz     @@6
                jmp     @@2
@@4:            inc     edx
                jz      @@3
                jmp     @@2
@@5:            inc     edx
                jz      @@2
@@3:            inc     eax
@@2:
end;

function zgte (a: zin; b: zin): boolean; assembler;
asm
                xor     eax,eax
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     edx,[edi+psign]
                cmp     edx,[esi+psign]
                je      @@1
                inc     edx
                jz      @@2
                jmp     @@3
@@1:            mov     ecx,[edi]
                cmp     ecx,[esi]
                jb      @@4
                ja      @@5
@@6:            mov     ebx,[edi+ecx*4]
                cmp     ebx,[esi+ecx*4]
                jb      @@4
                ja      @@5
                dec     ecx
                jnz     @@6
                jmp     @@3
@@4:            inc     edx
                jz      @@3
                jmp     @@2
@@5:            inc     edx
                jz      @@2
@@3:            inc     eax
@@2:
end;

procedure zchs (var a: zin); assembler;
asm
                mov     edi,[ebp+8]
                cmp     dword ptr [edi],1
                jne     @@1
                cmp     dword ptr [edi+4],0
                je      @@2
@@1:            mov     eax,[edi+psign]
                neg     eax
                mov     [edi+psign],eax
@@2:
end;

procedure zadd (a,b: zin; var c: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     edi,[ebp+12]
                mov     ebx,[ebp+16]
                mov     edx,[esi+psign]
                cmp     edx,[edi+psign]
                jne     @@10
                mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@11
                xchg    edi,esi
                xchg    edx,ecx
@@11:           mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@12
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@13:           mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@13
                jnc     @@12
                test    edx,edx
                jz      @@14
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@15:           mov     eax,[esi+edx*4]
                inc     eax
                jne     @@16
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@15
@@14:           mov     dword ptr [ebx],1
                mov     ebx,[ebp+16]
                inc     dword ptr [ebx]
                jmp     @@17
@@12:           test    edx,edx
                jz      @@18
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@19:           mov     eax,[esi+edx*4]
@@16:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@19
@@18:           mov     [ebx],ecx
                jmp     @@17
@@10:           mov     ecx,[esi]
                cmp     ecx,[edi]
                jb      @@20
                ja      @@21
@@22:           mov     eax,[esi+ecx*4]
                cmp     eax,[edi+ecx*4]
                jb      @@20
                ja      @@21
                dec     ecx
                jnz     @@22
                mov     dword ptr [ebx],1
                mov     dword ptr [ebx+4],0
                mov     dword ptr [ebx+psign],1
                jmp     @@17
@@20:           xchg    edi,esi
                neg     edx
@@21:           mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@23
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@24:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@24
                jnb     @@23
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@25:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@26
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@25
                jmp     @@27
@@23:           test    edx,edx
                jz      @@27
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@28:           mov     eax,[esi+edx*4]
@@26:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@28
@@27:           mov     ebx,[ebp+16]
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jne     @@17
                dec     ecx
                jz      @@17
@@29:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@30
                dec     ecx
                jnz     @@29
                inc     ecx
@@30:           mov     [ebx],ecx
@@17:
end;

procedure zadd_1 (a: zin; n: longword; var b: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                cmp     dword ptr [edi+psign],1
                je      @@10
                cmp     ecx,1
                ja      @@11
                cmp     [edi+4],ebx
                ja      @@11
@@12:           mov     dword ptr [esi],1
                sub     ebx,[edi+4]
                mov     [esi+4],ebx
                mov     dword ptr [esi+psign],1
                jmp     @@13
@@11:           mov     dword ptr [esi+psign],4294967295
                mov     edx,ecx
@@14:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@14
                mov     [esi],edx
                mov     eax,ebx
                add     esi,4
                sub     [esi],eax
                jnc     @@13
                dec     edx
@@15:           add     esi,4
                sbb     dword ptr[esi],1
                jnc     @@16
                dec     edx
                jnz     @@15
                jmp     @@13
@@16:           jnz     @@13
                dec     edx
                jnz     @@13
                mov     esi,[ebp+16]
                dec     dword ptr [esi]
                jmp     @@13
@@10:           mov     dword ptr [esi+psign],1
                mov     edx,ecx
@@18:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@18
                mov     eax,ebx
                add     [esi+4],eax
                jc      @@19
                mov     [esi],edx
                jmp     @@13
@@19:           add     esi,4
                mov     ecx,edx
                dec     ecx
                jz      @@20
@@21:           add     esi,4
                inc     dword ptr [esi]
                jnz     @@22
                dec     ecx
                jnz     @@21
@@20:           mov     dword ptr[esi+4],1
                inc     edx
@@22:           mov     esi,[ebp+16]
                mov     [esi],edx
@@13:
end;

procedure zsub (a,b: zin; var c: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     edi,[ebp+12]
                mov     ebx,[ebp+16]
                mov     edx,[esi+psign]
                cmp     edx,[edi+psign]
                je      @@10
                mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@11
                xchg    edi,esi
                xchg    edx,ecx
@@11:           mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@12
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@13:           mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@13
                jnc     @@12
                test    edx,edx
                jz      @@14
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@15:           mov     eax,[esi+edx*4]
                inc     eax
                jne     @@16
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@15
@@14:           mov     dword ptr [ebx],1
                mov     ebx,[ebp+16]
                inc     dword ptr [ebx]
                jmp     @@17
@@12:           test    edx,edx
                jz      @@18
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@19:           mov     eax,[esi+edx*4]
@@16:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@19
@@18:           mov     [ebx],ecx
                jmp     @@17
@@10:           mov     ecx,[esi]
                cmp     ecx,[edi]
                jb      @@20
                ja      @@21
@@22:           mov     eax,[esi+ecx*4]
                cmp     eax,[edi+ecx*4]
                jb      @@20
                ja      @@21
                dec     ecx
                jnz     @@22
                mov     dword ptr [ebx],1
                mov     dword ptr [ebx+4],0
                mov     dword ptr [ebx+psign],1
                jmp     @@17
@@20:           xchg    edi,esi
                neg     edx
@@21:           mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@23
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@24:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@24
                jnb     @@23
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@25:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@26
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@25
                jmp     @@27
@@23:           test    edx,edx
                jz      @@27
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@28:           mov     eax,[esi+edx*4]
@@26:           mov     [ebx+edx*4],eax
                inc     edx
                jne     @@28
@@27:           mov     ebx,[ebp+16]
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jne     @@17
                dec     ecx
                jz      @@17
@@29:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@30
                dec     ecx
                jnz     @@29
                inc     ecx
@@30:           mov     [ebx],ecx
@@17:
end;

procedure zsub_1 (a: zin; n: longword; var b: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                cmp     dword ptr [edi+psign],1
                jne     @@100
                cmp     ecx,1
                ja      @@110
                cmp     [edi+4],ebx
                jae     @@110
@@120:          mov     dword ptr [esi],1
                sub     ebx,[edi+4]
                mov     [esi+4],ebx
                mov     dword ptr [esi+psign],4294967295
                jmp     @@130
@@110:          mov     dword ptr [esi+psign],1
                mov     edx,ecx
@@140:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@140
                mov     [esi],edx
                mov     eax,ebx
                add     esi,4
                sub     [esi],eax
                jnc     @@130
                dec     edx
@@150:          add     esi,4
                sbb     dword ptr[esi],1
                jnc     @@160
                dec     edx
                jnz     @@150
                jmp     @@130
@@160:          jnz     @@130
                dec     edx
                jnz     @@130
                mov     esi,[ebp+16]
                dec     dword ptr [esi]
@@170:          jmp     @@130
@@100:          mov     dword ptr [esi+psign],4294967295
                mov     edx,ecx
@@180:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@180
                mov     eax,ebx
                add     [esi+4],eax
                jc      @@190
                mov     [esi],edx
                jmp     @@130
@@190:          add     esi,4
                mov     ecx,edx
                dec     ecx
                jz      @@200
@@210:          add     esi,4
                inc     dword ptr [esi]
                jnz     @@220
                dec     ecx
                jnz     @@210
@@200:          mov     dword ptr[esi+4],1
                inc     edx
@@220:          mov     esi,[ebp+16]
                mov     [esi],edx
@@130:
end;

procedure zmul (a,b: zin; var c: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                mov     edx,[edi]
                mov     ebx,[ebp+16]
                mov     eax,[esi+psign]
                cmp     eax,[edi+psign]
                je      @@10
                mov     dword ptr [ebx+psign],4294967295
                jne     @@11
@@10:           mov     dword ptr [ebx+psign],1
@@11:           cmp     ecx,edx
                jae     @@12
                xchg    edi,esi
                xchg    edx,ecx
@@12:           mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@13:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@13
                mov     [ebx],edx
                dec     loc_lb
                jz      @@14
@@15:           add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@16:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@16
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@15
@@14:           mov     esi,[ebp+16]
                mov     ecx,pm
@@17:           cmp     dword ptr[esi+ecx*4],0
                jne     @@18
                dec     ecx
                jnz     @@17
                inc     ecx
                mov     dword ptr [esi+psign],1
@@18:           mov     [esi],ecx
end;

procedure zmul_1 (a: zin; n: longword; var b: zin); assembler;
asm
                mov     esi,[ebp+16]
                mov     eax,[ebp+12]
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                cmp     eax,0
                jne     @@1
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],0
                mov     dword ptr [esi+psign],1
                jmp     @@3
@@1:            mov     ebx,[edi+psign]
                mov     [esi+psign],ebx
                mov     pm,eax
                mov     [esi],ecx
                add     edi,4
                add     esi,4
                xor     ebx,ebx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                neg     ecx
@@2:            mov     eax,[edi+ecx*4]
                mul     pm
                add     eax,ebx
                adc     edx,0
                mov     [esi+ecx*4],eax
                mov     ebx,edx
                inc     ecx
                jne     @@2
                cmp     ebx,0
                je      @@3
                mov     esi,[ebp+16]
                inc     dword ptr [esi]
                mov     ecx,[esi]
                mov     [esi+ecx*4],ebx
@@3:
end;

procedure zsqr (a: zin; var b: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     ebx,[ebp+12]
                mov     dword ptr [ebx+psign],1
                mov     edi,esi
                mov     edx,ecx
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@10:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@10
                mov     [ebx],edx
                dec     loc_lb
                jz      @@11
@@12:           add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@13:           mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@13
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@12
@@11:           mov     esi,[ebp+12]
                mov     ecx,pm
@@14:           cmp     dword ptr[esi+ecx*4],0
                jne     @@15
                dec     ecx
                jnz     @@14
                inc     ecx
@@15:           mov     [esi],ecx
@@16:
end;

procedure zediv (a,b: zin; var c,d: zin); assembler;
asm
                mov     esi,[ebp+20]
                mov     dword ptr [esi+psign],1
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@11
                mov     edi,[ebp+16]
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                mov     esi,[ebp+16]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@15
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@16
                dec     ecx
                jnz     @@16
                inc     ecx
@@16:           mov     [esi],ecx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                mov     esi,[ebp+16]
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@17:           sub     esi,4
                cmp     edx,reg_b
                je      @@18
                div     reg_b
                jmp     @@19
@@18:           add     edx,eax
                mov     eax,4294967295
                jnc     @@19
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@20
                mov     ebx,[edi]
@@20:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@21
@@19:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@22
                mov     ebx,[edi]
@@22:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@23
                ja      @@24
                test    eax,eax
                jz      @@23
@@24:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@23
                ja      @@24
                test    eax,eax
                jnz     @@24
@@23:           neg     eax
                adc     edx,0
                neg     edx
@@21:           popf
                jnz     @@17
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@25
                mov     dword ptr [edi],1
                jmp     @@26
@@25:           mov     [edi+8],edx
                mov     dword ptr [edi],2
@@26:           mov     edi,[ebp+16]
                mov     ecx,la1
@@27:           cmp     dword ptr [edi+ecx*4],0
                jne     @@28
                dec     ecx
                jnz     @@27
                inc     ecx
@@28:           mov     [edi],ecx
                jmp     @@12
@@13:           mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
                mov     ebx,[ebp+16]
@@29:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@29
@@30:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@31
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@32
@@31:           mov     ebx,[ebp+16]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@33
@@32:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@34:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@35
                div     reg_a
                mov     reg_q,eax
                jmp     @@36
@@35:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@37
@@36:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@37
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@37
                dec     reg_q
@@37:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@38:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@38
                cmp     [esi],ebx
                je      @@39
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@40:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@40
@@39:           mov     ebx,[ebp+16]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@34
@@33:           mov     ecx,la1
@@41:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@42
                dec     ecx
                jnz     @@41
                inc     ecx
@@42:           mov     [ebx],ecx
@@43:           mov     esi,[ebp+20]
                mov     ecx,la2
@@44:           cmp     dword ptr[esi+ecx*4],0
                jne     @@45
                dec     ecx
                jnz     @@44
                inc     ecx
@@45:           mov     [esi],ecx
@@12:           mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     esi,[ebp+16]
                mov     eax,[ebx+psign]
                cmp     dword ptr [edi+psign],1
                jne     @@46
                cmp     dword ptr [esi+4],0
                jne     @@47
                cmp     dword ptr [esi],1
                jne     @@47
                mov     eax,1
@@47:           mov     [esi+psign],eax
                jmp     @@48
@@46:           neg     eax
                mov     [esi+psign],eax
                mov     edi,[ebp+20]
                cmp     dword ptr [edi],1
                jne     @@49
                cmp     dword ptr [edi+4],0
                je      @@48
@@49:           mov     ecx,[esi]
                mov     eax,1
                add     esi,4
                add     [esi],eax
                jnc     @@50
                dec     ecx
                jz      @@51
@@52:           add     esi,4
                inc     dword ptr [esi]
                jnz     @@50
                dec     ecx
                jnz     @@52
@@51:           add     esi,4
                mov     dword ptr[esi],1
                mov     esi,[ebp+16]
                inc     dword ptr [esi]
@@50:           mov     esi,[ebp+12]
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                sub     edx,ecx
                jecxz   @@53
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                neg     ecx
                clc
@@54:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [edi+ecx*4],eax
                inc     ecx
                jne     @@54
                jnb     @@53
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@55:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@56
                mov     [edi+edx*4],eax
                inc     edx
                jne     @@55
                jmp     @@57
@@53:           test    edx,edx
                jz      @@57
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@58:           mov     eax,[esi+edx*4]
@@56:           mov     [edi+edx*4],eax
                inc     edx
                jne     @@58
@@57:           mov     esi,[ebp+12]
                mov     ecx,[esi]
                mov     edi,[ebp+20]
@@59:           cmp     dword ptr[edi+ecx*4],0
                jne     @@60
                dec     ecx
                jnz     @@59
                inc     ecx
@@60:           mov     [edi],ecx
@@48:
end;

procedure zequo (a,b: zin; var c: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                lea     esi,natz1
                mov     [esi],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@11
                mov     edi,[ebp+16]
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                mov     esi,[ebp+16]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@15
                lea     ebx,natz1
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@16
                dec     ecx
                jnz     @@16
                inc     ecx
@@16:           mov     [esi],ecx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                mov     esi,[ebp+16]
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@17:           sub     esi,4
                cmp     edx,reg_b
                je      @@18
                div     reg_b
                jmp     @@19
@@18:           add     edx,eax
                mov     eax,4294967295
                jnc     @@19
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@20
                mov     ebx,[edi]
@@20:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@21
@@19:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@22
                mov     ebx,[edi]
@@22:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@23
                ja      @@24
                test    eax,eax
                jz      @@23
@@24:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@23
                ja      @@24
                test    eax,eax
                jnz     @@24
@@23:           neg     eax
                adc     edx,0
                neg     edx
@@21:           popf
                jnz     @@17
                shrd    eax,edx,cl
                shr     edx,cl
                lea     edi,natz1
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@25
                mov     dword ptr [edi],1
                jmp     @@26
@@25:           mov     [edi+8],edx
                mov     dword ptr [edi],2
@@26:           mov     edi,[ebp+16]
                mov     ecx,la1
@@27:           cmp     dword ptr [edi+ecx*4],0
                jne     @@28
                dec     ecx
                jnz     @@27
                inc     ecx
@@28:           mov     [edi],ecx
                jmp     @@12
@@13:           mov     edx,la1
                lea     esi,natz1
                mov     [esi],ecx
                mov     ebx,[ebp+16]
@@29:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@29
@@30:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@31
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@32
@@31:           mov     ebx,[ebp+16]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@33
@@32:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@34:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@35
                div     reg_a
                mov     reg_q,eax
                jmp     @@36
@@35:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@37
@@36:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@37
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@37
                dec     reg_q
@@37:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@38:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@38
                cmp     [esi],ebx
                je      @@39
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@40:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@40
@@39:           mov     ebx,[ebp+16]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@34
@@33:           mov     ecx,la1
@@41:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@42
                dec     ecx
                jnz     @@41
                inc     ecx
@@42:           mov     [ebx],ecx
@@43:           lea     esi,natz1
                mov     ecx,la2
@@44:           cmp     dword ptr[esi+ecx*4],0
                jne     @@45
                dec     ecx
                jnz     @@44
                inc     ecx
@@45:           mov     [esi],ecx
@@12:           mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ebx,[ebp+12]
                mov     eax,[ebx+psign]
                cmp     dword ptr [edi+psign],1
                jne     @@46
                cmp     dword ptr [esi+4],0
                jne     @@47
                cmp     dword ptr [esi],1
                jne     @@47
                mov     eax,1
@@47:           mov     [esi+psign],eax
                jmp     @@48
@@46:           neg     eax
                mov     [esi+psign],eax
                lea     ebx,natz1
                cmp     dword ptr [ebx+4],0
                jne     @@49
                cmp     dword ptr [ebx],1
                je      @@48
@@49:           mov     ecx,[esi]
                mov     eax,1
                add     esi,4
                add     [esi],eax
                jnc     @@48
                dec     ecx
                jz      @@50
@@51:           add     esi,4
                inc     dword ptr [esi]
                jnz     @@48
                dec     ecx
                jnz     @@51
@@50:           add     esi,4
                mov     dword ptr[esi],1
                mov     esi,[ebp+16]
                inc     dword ptr [esi]
@@48:
end;

procedure zequo_1 (a: zin; n: longword; var b: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                xor     edx,edx
@@10:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@10
                mov     ecx,[edi]
                cmp     dword ptr[esi+ecx*4],0
                jne     @@11
                dec     ecx
                jnz     @@11
                inc     ecx
@@11:           mov     [esi],ecx
                cmp     dword ptr [edi+psign],1
                jne     @@12
                mov     dword ptr [esi+psign],1
                jmp     @@13
@@12:           mov     dword ptr [esi+psign],4294967295
                cmp     edx,0
                je      @@13
                mov     ecx,[esi]
                mov     eax,1
                add     esi,4
                add     [esi],eax
                jnc     @@13
                dec     ecx
                jz      @@14
@@15:           add     esi,4
                inc     dword ptr [esi]
                jnz     @@13
                dec     ecx
                jnz     @@15
@@14:           add     esi,4
                mov     dword ptr[esi],1
                mov     esi,[ebp+8]
                inc     dword ptr [esi]
@@13:
end;

procedure zemod (a,b: zin; var c: zin); assembler;
asm
                mov     ebx,[ebp+16];
                mov     dword ptr [ebx+psign],1
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     [ebx],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@11
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@15
                mov     ebx,[ebp+16]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@16:           cmp     edx,reg_b
                je      @@17
                div     reg_b
                jmp     @@18
@@17:           add     edx,eax
                mov     eax,4294967295
                jnc     @@18
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@19
                mov     ebx,[edi]
@@19:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@20
@@18:           mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@21
                mov     ebx,[edi]
@@21:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@22
                ja      @@23
                test    eax,eax
                jz      @@22
@@23:           dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@22
                ja      @@23
                test    eax,eax
                jnz     @@23
@@22:           neg     eax
                adc     edx,0
                neg     edx
@@20:           popf
                jnz     @@16
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+16]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@24
                mov     dword ptr [edi],1
                jmp     @@12
@@24:           mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@12
@@13:           mov     edx,la1
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@25:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@25
@@26:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@27
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@28
@@27:           test    edx,edx
                jz      @@29
@@28:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@30:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@31
                div     reg_a
                mov     reg_q,eax
                jmp     @@32
@@31:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@33
@@32:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@33
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@33
                dec     reg_q
@@33:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@34:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@34
                cmp     [esi],ebx
                je      @@35
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@36:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@36
@@35:           dec     reg_c
                jne     @@30
@@29:           mov     esi,[ebp+16]
                mov     ecx,la2
@@37:           cmp     dword ptr[esi+ecx*4],0
                jne     @@38
                dec     ecx
                jnz     @@37
                inc     ecx
@@38:           mov     [esi],ecx
@@12:           mov     esi,[ebp+8]
                cmp     dword ptr [esi+psign],1
                je      @@39
                mov     edi,[ebp+16]
                cmp     dword ptr [edi],1
                jne     @@40
                cmp     dword ptr [edi+4],0
                jne     @@40
                jmp     @@39
@@40:           mov     esi,[ebp+12]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+16]
                mov     ecx,[edi]
                add     edi,4
                sub     edx,ecx
                jecxz   @@41
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                neg     ecx
                clc
@@42:           mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [edi+ecx*4],eax
                inc     ecx
                jne     @@42
                jnb     @@41
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@43:           mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@44
                mov     [edi+edx*4],eax
                inc     edx
                jne     @@43
                jmp     @@45
@@41:           test    edx,edx
                jz      @@45
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@46:           mov     eax,[esi+edx*4]
@@44:           mov     [edi+edx*4],eax
                inc     edx
                jne     @@46
@@45:           mov     esi,[ebp+12]
                mov     ecx,[esi]
                mov     edi,[ebp+16]
                cmp     dword ptr[edi+ecx*4],0
                jne     @@47
                dec     ecx
                jz      @@48
@@49:           cmp     dword ptr[edi+ecx*4],0
                jne     @@47
                dec     ecx
                jnz     @@49
@@48:           inc     ecx
@@47:           mov     [edi],ecx
@@39:
end;

procedure zemod_1 (a: zin; n: longword; var r: longword); assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                mov     esi,[ebp+16]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1
                cmp     dword ptr [edi+psign],1
                je      @@2
                cmp     edx,0
                je      @@2
                sub     ebx,edx
                mov     edx,ebx
@@2:            mov     [esi],edx
end;

procedure zdiv (a,b: zin; var c,d: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@11
                mov     edi,[ebp+16]
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                mov     dword ptr [edi+psign],1
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                mov     esi,[ebp+16]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@15
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@16
                dec     ecx
                jnz     @@16
                inc     ecx
@@16:           mov     [esi],ecx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                mov     esi,[ebp+16]
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@17:           sub     esi,4
                cmp     edx,reg_b
                je      @@18
                div     reg_b
                jmp     @@19
@@18:           add     edx,eax
                mov     eax,4294967295
                jnc     @@19
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@20
                mov     ebx,[edi]
@@20:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@21
@@19:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@22
                mov     ebx,[edi]
@@22:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@23
                ja      @@24
                test    eax,eax
                jz      @@23
@@24:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@23
                ja      @@24
                test    eax,eax
                jnz     @@24
@@23:           neg     eax
                adc     edx,0
                neg     edx
@@21:           popf
                jnz     @@17
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@25
                mov     dword ptr [edi],1
                jmp     @@26
@@25:           mov     [edi+8],edx
                mov     dword ptr [edi],2
@@26:           mov     edi,[ebp+16]
                mov     ecx,la1
@@27:           cmp     dword ptr [edi+ecx*4],0
                jne     @@28
                dec     ecx
                jnz     @@27
                inc     ecx
@@28:           mov     [edi],ecx
                jmp     @@12
@@13:           mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
                mov     ebx,[ebp+16]
@@29:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@29
@@30:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@31
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@32
@@31:           mov     ebx,[ebp+16]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@33
@@32:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@34:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@35
                div     reg_a
                mov     reg_q,eax
                jmp     @@36
@@35:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@37
@@36:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@37
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@37
                dec     reg_q
@@37:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@38:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@38
                cmp     [esi],ebx
                je      @@39
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@40:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@40
@@39:           mov     ebx,[ebp+16]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@34
@@33:           mov     ecx,la1
@@41:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@42
                dec     ecx
                jnz     @@41
                inc     ecx
@@42:           mov     [ebx],ecx
@@43:           mov     esi,[ebp+20]
                mov     ecx,la2
@@44:           cmp     dword ptr[esi+ecx*4],0
                jne     @@45
                dec     ecx
                jnz     @@44
                inc     ecx
@@45:           mov     [esi],ecx
@@12:           mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi+psign]
                mov     eax,1
                mov     ebx,[ebp+16]
                cmp     dword ptr [ebx+4],0
                jne     @@46
                cmp     dword ptr [ebx],1
                je      @@47
@@46:           cmp     ecx,[esi+psign]
                je      @@47
                neg     eax
@@47:           mov     [ebx+psign],eax
                mov     ebx,[ebp+20]
                cmp     dword ptr [ebx+4],0
                jne     @@48
                cmp     dword ptr [ebx],1
                jne     @@48
                mov     ecx,1
@@48:           mov     [ebx+psign],ecx
end;

procedure zquo (a,b: zin; var c: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
@@11:           mov     edi,[ebp+16]
                mov     dword ptr [edi],1
                mov     dword ptr [edi+4],0
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                mov     esi,[ebp+16]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@15
                mov     ecx,la1
                cmp     dword ptr[esi+ecx*4],0
                jne     @@16
                dec     ecx
                jnz     @@16
                inc     ecx
@@16:           mov     [esi],ecx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                mov     esi,[ebp+16]
                add     edi,4
                add     esi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                lea     esi,[esi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     [esi],edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@17:           sub     esi,4
                cmp     edx,reg_b
                je      @@18
                div     reg_b
                jmp     @@19
@@18:           add     edx,eax
                mov     eax,4294967295
                jnc     @@19
                sub     edi,4
                mov     [esi],eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@20
                mov     ebx,[edi]
@@20:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@21
@@19:           mov     [esi],eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@22
                mov     ebx,[edi]
@@22:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     dword ptr [esi]
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@23
                ja      @@24
                test    eax,eax
                jz      @@23
@@24:           dec     dword ptr [esi]
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@23
                ja      @@24
                test    eax,eax
                jnz     @@24
@@23:           neg     eax
                adc     edx,0
                neg     edx
@@21:           popf
                jnz     @@17
                mov     edi,[ebp+16]
                mov     ecx,la1
@@25:           cmp     dword ptr [edi+ecx*4],0
                jne     @@26
                dec     ecx
                jnz     @@25
                inc     ecx
@@26:           mov     [edi],ecx
                jmp     @@12
@@13:           mov     edx,la1
                lea     esi,natz1
                mov     ebx,[ebp+16]
@@27:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     dword ptr [ebx+ecx*4],0
                dec     ecx
                jnz     @@27
@@28:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                add     edi,4
                lea     edi,[edi+ecx*4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@29
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@30
@@29:           mov     ebx,[ebp+16]
                mov     dword ptr [ebx+edx*4],0
                test    edx,edx
                jz      @@31
@@30:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@32:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@33
                div     reg_a
                mov     reg_q,eax
                jmp     @@34
@@33:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@35
@@34:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@35
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@35
                dec     reg_q
@@35:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@36:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@36
                cmp     [esi],ebx
                je      @@37
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@38:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@38
@@37:           mov     ebx,[ebp+16]
                mov     edx,reg_c
                mov     eax,reg_q
                mov     [ebx+edx*4],eax
                dec     reg_c
                jne     @@32
@@31:           mov     ecx,la1
@@39:           cmp     dword ptr[ebx+ecx*4],0
                jne     @@40
                dec     ecx
                jnz     @@39
                inc     ecx
@@40:           mov     [ebx],ecx
@@12:           mov     ebx,[ebp+16]
                cmp     dword ptr [ebx+4],0
                jne     @@41
                cmp     dword ptr [ebx],1
                jne     @@41
                mov     dword ptr [ebx+psign],1
                jmp     @@42
@@41:           mov     eax,1
                mov     edi,[ebp+8]
                mov     esi,[ebp+12]
                mov     ecx,[edi+psign]
                cmp     ecx,[esi+psign]
                je      @@43
                neg     eax
@@43:           mov     [ebx+psign],eax
@@42:
end;

procedure zquo_1 (a: zin; n: longword; var b:zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@1
                mov     ecx,[edi]
                cmp     dword ptr[esi+ecx*4],0
                jne     @@2
                dec     ecx
                jnz     @@2
                inc     ecx
@@2:            mov     [esi],ecx
                cmp     dword ptr [esi+4],0
                jne     @@3
                cmp     dword ptr [esi],1
                je      @@4
@@3:            cmp     edx,0
                jne     @@5
@@4:            mov     dword ptr [esi+psign],1
                jmp     @@6
@@5:            mov     eax,[edi+psign]
                mov     [esi+psign],eax
@@6:
end;

procedure zrem (a,b: zin; var c: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@10
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@11:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@11
                jmp     @@12
@@10:           mov     la2,edx
                cmp     edx,2
                ja      @@13
                je      @@14
                mov     ebx,[esi+4]
                xor     edx,edx
@@15:           mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@15
                mov     ebx,[ebp+16]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@12
@@14:           mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@16:           cmp     edx,reg_b
                je      @@17
                div     reg_b
                jmp     @@18
@@17:           add     edx,eax
                mov     eax,4294967295
                jnc     @@18
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@19
                mov     ebx,[edi]
@@19:           add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@20
@@18:           mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@21
                mov     ebx,[edi]
@@21:           shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@22
                ja      @@23
                test    eax,eax
                jz      @@22
@@23:           dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@22
                ja      @@23
                test    eax,eax
                jnz     @@23
@@22:           neg     eax
                adc     edx,0
                neg     edx
@@20:           popf
                jnz     @@16
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+16]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@24
                mov     dword ptr [edi],1
                jmp     @@12
@@24:           mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@12
@@13:           mov     edx,la1
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@25:           mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@25
@@26:           lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@27
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@28
@@27:           test    edx,edx
                jz      @@29
@@28:           mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@30:           sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@31
                div     reg_a
                mov     reg_q,eax
                jmp     @@32
@@31:           add     edx,eax
                mov     reg_q,4294967295
                jc      @@33
@@32:           mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@33
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@33
                dec     reg_q
@@33:           mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@34:           mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@34
                cmp     [esi],ebx
                je      @@35
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@36:           mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@36
@@35:           dec     reg_c
                jne     @@30
@@29:           mov     esi,[ebp+16]
                mov     ecx,la2
@@37:           cmp     dword ptr[esi+ecx*4],0
                jne     @@38
                dec     ecx
                jnz     @@37
                inc     ecx
@@38:           mov     [esi],ecx
@@12:           mov     ebx,[ebp+16]
                cmp     dword ptr [ebx+4],0
                jne     @@39
                cmp     dword ptr [ebx],1
                jne     @@39
                mov     dword ptr [ebx+psign],1
                jmp     @@40
@@39:           mov     edi,[ebp+8]
                mov     eax,[edi+psign]
                mov     [ebx+psign],eax
@@40:
end;

procedure zrem_1 (a: zin; n: longword; var b: zin); assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                mov     esi,[ebp+16]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1
                mov     dword ptr [esi],1
                mov     [esi+4],edx
                cmp     edx,0
                jne     @@2
                mov     dword ptr [esi+psign],1
                jmp     @@3
@@2:            mov     eax,[edi+psign]
                mov     [esi+psign],eax
@@3:
end;

function zmod0 (a: zin; n: longword): longword; assembler;
asm
                mov     edi,[ebp+8]
                mov     ebx,[ebp+12]
                mov     ecx,[edi]
                xor     edx,edx
@@1:            mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@1
                mov     eax,edx
end;

procedure zpow_1 (a: zin; n: longword; var b: zin); assembler;
asm
                mov     esi,[ebp+16]
                mov     edx,[ebp+12]
                dec     edx
                jns     @@100
                mov     dword ptr [esi],1
                mov     dword ptr [esi+4],1
                mov     dword ptr [esi+psign],1
                jmp     @@110
@@100:          mov     edi,[ebp+8]
                lea     ebx,natz3
                mov     ecx,[edi]
                mov     [esi],ecx
                mov     [ebx],ecx
@@120:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@120
                mov     kz,edx
@@130:          test    kz,1
                jz      @@140
                mov     esi,[ebp+16]
                mov     ecx,[esi]
                lea     edi,natz3
                mov     edx,[edi]
                lea     ebx,natz1
                cmp     ecx,edx
                jae     @@150
                xchg    edi,esi
                xchg    edx,ecx
@@150:          mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@160:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@160
                mov     [ebx],edx
                dec     loc_lb
                jz      @@170
@@180:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@190:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@190
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@180
@@170:          lea     esi,natz1
                mov     edi,[ebp+16]
                mov     ecx,pm
@@200:          cmp     dword ptr[esi+ecx*4],0
                jne     @@210
                dec     ecx
                jnz     @@200
                inc     ecx
@@210:          mov     [edi],ecx
@@220:          mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@220
@@140:          shr     kz,1
                jz      @@110
                lea     esi,natz3
                mov     ecx,[esi]
                mov     edi,esi
                mov     edx,ecx
                lea     ebx,natz2
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@230:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@230
                mov     [ebx],edx
                dec     loc_lb
                jz      @@240
@@250:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@260:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@260
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@250
@@240:          lea     esi,natz2
                lea     edi,natz3
                mov     ecx,pm
@@270:          cmp     dword ptr[esi+ecx*4],0
                jne     @@280
                dec     ecx
                jnz     @@270
                inc     ecx
@@280:          mov     [edi],ecx
@@290:          mov     eax,[esi+ecx*4]
                mov     [edi+ecx*4],eax
                dec     ecx
                jnz     @@290
                jmp     @@130
@@110:          mov     edi,[ebp+8]
                mov     esi,[ebp+16]
                mov     eax,[edi+psign]
                cmp     eax,1
                je      @@112
                test    dword ptr [ebp+12],1
                jnz     @@112
                neg     eax
@@112:          mov     [esi+psign],eax
end;

PROCEDURE zaddmod (a,b,c: zin; var d: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     edi,[ebp+12]
                lea     ebx,natz3
                mov     edx,[esi+psign]
                cmp     edx,[edi+psign]
                jne     @@100
                mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@110
                xchg    edi,esi
                xchg    edx,ecx
@@110:          mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@120
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@130:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@130
                jnc     @@120
                test    edx,edx
                jz      @@140
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@150:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@160
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@150
@@140:          mov     dword ptr [ebx],1
                lea     ebx,natz3
                inc     dword ptr [ebx]
                jmp     @@170
@@120:          test    edx,edx
                jz      @@180
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@190:          mov     eax,[esi+edx*4]
@@160:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@190
@@180:          mov     [ebx],ecx
                jmp     @@170
@@100:          mov     ecx,[esi]
                cmp     ecx,[edi]
                jb      @@200
                ja      @@210
@@220:          mov     eax,[esi+ecx*4]
                cmp     eax,[edi+ecx*4]
                jb      @@200
                ja      @@210
                dec     ecx
                jnz     @@220
                mov     dword ptr [ebx],1
                mov     dword ptr [ebx+4],0
                mov     dword ptr [ebx+psign],1
                jmp     @@170
@@200:          xchg    edi,esi
                neg     edx
@@210:          mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@230
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@240:          mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@240
                jnb     @@230
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@250:          mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@260
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@250
                jmp     @@270
@@230:          test    edx,edx
                jz      @@270
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@280:          mov     eax,[esi+edx*4]
@@260:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@280
@@270:          lea     ebx,natz3
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jne     @@170
                dec     ecx
                jz      @@170
@@290:          cmp     dword ptr[ebx+ecx*4],0
                jne     @@300
                dec     ecx
                jnz     @@290
                inc     ecx
@@300:          mov     [ebx],ecx
@@170:          lea     edi,natz3
                mov     ecx,[edi]
                mov     ebx,[ebp+20];
                mov     dword ptr [ebx+psign],1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@310
                mov     [ebx],ecx
@@320:          mov     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@320
                jmp     @@330
@@310:          mov     la2,edx
                cmp     edx,2
                ja      @@340
                je      @@350
                mov     ebx,[esi+4]
                xor     edx,edx
@@360:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@360
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@330
@@350:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@370:          cmp     edx,reg_b
                je      @@380
                div     reg_b
                jmp     @@390
@@380:          add     edx,eax
                mov     eax,4294967295
                jnc     @@390
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@400
                mov     ebx,[edi]
@@400:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@410
@@390:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@420
                mov     ebx,[edi]
@@420:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@430
                ja      @@440
                test    eax,eax
                jz      @@430
@@440:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@430
                ja      @@440
                test    eax,eax
                jnz     @@440
@@430:          neg     eax
                adc     edx,0
                neg     edx
@@410:          popf
                jnz     @@370
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@450
                mov     dword ptr [edi],1
                jmp     @@330
@@450:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@330
@@340:          mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@460:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@460
@@470:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@480
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@490
@@480:          test    edx,edx
                jz      @@500
@@490:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@510:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@520
                div     reg_a
                mov     reg_q,eax
                jmp     @@530
@@520:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@540
@@530:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@540
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@540
                dec     reg_q
@@540:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@550:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@550
                cmp     [esi],ebx
                je      @@560
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@570:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@570
@@560:          dec     reg_c
                jne     @@510
@@500:          mov     esi,[ebp+20]
                mov     ecx,la2
@@580:          cmp     dword ptr[esi+ecx*4],0
                jne     @@590
                dec     ecx
                jnz     @@580
                inc     ecx
@@590:          mov     [esi],ecx
@@330:          lea     ebx,natz3
                cmp     dword ptr [ebx+psign],1
                je      @@600
                mov     esi,[ebp+16]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+20]
                mov     ecx,[edi]
                add     edi,4
                sub     edx,ecx
                jecxz   @@610
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                neg     ecx
                clc
@@620:          mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [edi+ecx*4],eax
                inc     ecx
                jne     @@620
                jnb     @@610
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@630:          mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@640
                mov     [edi+edx*4],eax
                inc     edx
                jne     @@630
                jmp     @@650
@@610:          test    edx,edx
                jz      @@650
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@660:          mov     eax,[esi+edx*4]
@@640:          mov     [edi+edx*4],eax
                inc     edx
                jne     @@660
@@650:          mov     esi,[ebp+16]
                mov     ecx,[esi]
                mov     edi,[ebp+20]
                cmp     dword ptr[edi+ecx*4],0
                jnz     @@670
                dec     ecx
                jz      @@680
@@690:          cmp     dword ptr[edi+ecx*4],0
                jne     @@670
                dec     ecx
                jnz     @@690
@@680:          inc     ecx
@@670:          mov     [edi],ecx
@@600:
end;

PROCEDURE zsubmod (a,b,c: zin; var d: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     edi,[ebp+12]
                lea     ebx,natz3
                mov     edx,[esi+psign]
                cmp     edx,[edi+psign]
                je      @@100
                mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                cmp     edx,ecx
                jae     @@110
                xchg    edi,esi
                xchg    edx,ecx
@@110:          mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@120
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@130:          mov     eax,[esi+ecx*4]
                adc     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@130
                jnc     @@120
                test    edx,edx
                jz      @@140
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@150:          mov     eax,[esi+edx*4]
                inc     eax
                jne     @@160
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@150
@@140:          mov     dword ptr [ebx],1
                lea     ebx,natz3
                inc     dword ptr [ebx]
                jmp     @@170
@@120:          test    edx,edx
                jz      @@180
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@190:          mov     eax,[esi+edx*4]
@@160:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@190
@@180:          mov     [ebx],ecx
                jmp     @@170
@@100:          mov     ecx,[esi]
                cmp     ecx,[edi]
                jb      @@200
                ja      @@210
@@220:          mov     eax,[esi+ecx*4]
                cmp     eax,[edi+ecx*4]
                jb      @@200
                ja      @@210
                dec     ecx
                jnz     @@220
                mov     dword ptr [ebx],1
                mov     dword ptr [ebx+4],0
                mov     dword ptr [ebx+psign],1
                jmp     @@170
@@200:          xchg    edi,esi
                neg     edx
@@210:          mov     [ebx+psign],edx
                mov     edx,[esi]
                add     esi,4
                mov     ecx,[edi]
                add     edi,4
                mov     [ebx],edx
                add     ebx,4
                sub     edx,ecx
                jecxz   @@230
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                lea     ebx,[ebx+ecx*4]
                neg     ecx
                clc
@@240:          mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                inc     ecx
                jne     @@240
                jnb     @@230
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@250:          mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@260
                mov     [ebx+edx*4],eax
                inc     edx
                jne     @@250
                jmp     @@270
@@230:          test    edx,edx
                jz      @@270
                lea     esi,[esi+edx*4]
                lea     ebx,[ebx+edx*4]
                neg     edx
@@280:          mov     eax,[esi+edx*4]
@@260:          mov     [ebx+edx*4],eax
                inc     edx
                jne     @@280
@@270:          lea     ebx,natz3
                mov     ecx,[ebx]
                cmp     dword ptr[ebx+ecx*4],0
                jnz     @@170
                dec     ecx
                jz      @@170
@@290:          cmp     dword ptr[ebx+ecx*4],0
                jne     @@300
                dec     ecx
                jnz     @@290
                inc     ecx
@@300:          mov     [ebx],ecx
@@170:          lea     edi,natz3
                mov     ecx,[edi]
                mov     ebx,[ebp+20];
                mov     dword ptr [ebx+psign],1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@310
                mov     [ebx],ecx
@@320:          mov     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@320
                jmp     @@330
@@310:          mov     la2,edx
                cmp     edx,2
                ja      @@340
                je      @@350
                mov     ebx,[esi+4]
                xor     edx,edx
@@360:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@360
                mov     ebx,[ebp+20]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@330
@@350:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@370:          cmp     edx,reg_b
                je      @@380
                div     reg_b
                jmp     @@390
@@380:          add     edx,eax
                mov     eax,4294967295
                jnc     @@390
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@400
                mov     ebx,[edi]
@@400:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@410
@@390:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@420
                mov     ebx,[edi]
@@420:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@430
                ja      @@440
                test    eax,eax
                jz      @@430
@@440:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@430
                ja      @@440
                test    eax,eax
                jnz     @@440
@@430:          neg     eax
                adc     edx,0
                neg     edx
@@410:          popf
                jnz     @@370
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+20]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@450
                mov     dword ptr [edi],1
                jmp     @@330
@@450:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@330
@@340:          mov     edx,la1
                mov     esi,[ebp+20]
                mov     [esi],ecx
@@460:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@460
@@470:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@480
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@490
@@480:          test    edx,edx
                jz      @@500
@@490:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@510:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@520
                div     reg_a
                mov     reg_q,eax
                jmp     @@530
@@520:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@540
@@530:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@540
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@540
                dec     reg_q
@@540:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@550:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@550
                cmp     [esi],ebx
                je      @@560
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@570:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@570
@@560:          dec     reg_c
                jne     @@510
@@500:          mov     esi,[ebp+20]
                mov     ecx,la2
@@580:          cmp     dword ptr[esi+ecx*4],0
                jne     @@590
                dec     ecx
                jnz     @@580
                inc     ecx
@@590:          mov     [esi],ecx
@@330:          lea     ebx,natz3
                cmp     dword ptr [ebx+psign],1
                je      @@600
                mov     esi,[ebp+16]
                mov     edx,[esi]
                add     esi,4
                mov     edi,[ebp+20]
                mov     ecx,[edi]
                add     edi,4
                sub     edx,ecx
                jecxz   @@610
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                neg     ecx
                clc
@@620:          mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [edi+ecx*4],eax
                inc     ecx
                jne     @@620
                jnb     @@610
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@630:          mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@640
                mov     [edi+edx*4],eax
                inc     edx
                jne     @@630
                jmp     @@650
@@610:          test    edx,edx
                jz      @@650
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@660:          mov     eax,[esi+edx*4]
@@640:          mov     [edi+edx*4],eax
                inc     edx
                jne     @@660
@@650:          mov     esi,[ebp+16]
                mov     ecx,[esi]
                mov     edi,[ebp+20]
                cmp     dword ptr[edi+ecx*4],0
                jnz     @@670
                dec     ecx
                jz      @@680
@@690:          cmp     dword ptr[edi+ecx*4],0
                jne     @@670
                dec     ecx
                jnz     @@690
@@680:          inc     ecx
@@670:          mov     [edi],ecx
@@600:
end;

PROCEDURE zmulmod (a,b,c: zin; var d: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                mov     edi,[ebp+12]
                mov     edx,[edi]
                lea     ebx,natz3
                cmp     ecx,edx
                jae     @@100
                xchg    edi,esi
                xchg    edx,ecx
@@100:          mov     pm,ecx
                add     pm,edx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@110:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@110
                mov     [ebx],edx
                dec     loc_lb
                jz      @@120
@@130:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@140:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@140
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@130
@@120:          lea     edi,natz3
                mov     ecx,pm
@@150:          cmp     dword ptr[edi+ecx*4],0
                jne     @@160
                dec     ecx
                jnz     @@150
                inc     ecx
@@160:          mov     [edi],ecx
@@170:          lea     ebx,natz1
                mov     ecx,[edi]
                mov     la1,ecx
                mov     esi,[ebp+16]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@180
                mov     [ebx],ecx
@@190:          mov     eax,[edi+ecx*4]
                mov     [ebx+ecx*4],eax
                dec     ecx
                jnz     @@190
                jmp     @@200
@@180:          mov     la2,edx
                cmp     edx,2
                ja      @@210
                je      @@220
                mov     ebx,[esi+4]
                xor     edx,edx
@@230:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@230
                lea     ebx,natz1
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@200
@@220:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@240:          cmp     edx,reg_b
                je      @@250
                div     reg_b
                jmp     @@260
@@250:          add     edx,eax
                mov     eax,4294967295
                jnc     @@260
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@270
                mov     ebx,[edi]
@@270:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@280
@@260:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@290
                mov     ebx,[edi]
@@290:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@300
                ja      @@310
                test    eax,eax
                jz      @@300
@@310:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@300
                ja      @@310
                test    eax,eax
                jnz     @@310
@@300:          neg     eax
                adc     edx,0
                neg     edx
@@280:          popf
                jnz     @@240
                shrd    eax,edx,cl
                shr     edx,cl
                lea     edi,natz1
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@320
                mov     dword ptr [edi],1
                jmp     @@200
@@320:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@200
@@210:          mov     edx,la1
                lea     esi,natz1
                mov     [esi],ecx
@@330:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@330
@@340:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+16]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@350
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@360
@@350:          test    edx,edx
                jz      @@370
@@360:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@380:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@390
                div     reg_a
                mov     reg_q,eax
                jmp     @@400
@@390:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@410
@@400:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@410
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@410
                dec     reg_q
@@410:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@420:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@420
                cmp     [esi],ebx
                je      @@430
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@440:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@440
@@430:          dec     reg_c
                jne     @@380
@@370:          lea     esi,natz1
                mov     ecx,la2
@@450:          cmp     dword ptr[esi+ecx*4],0
                jne     @@460
                dec     ecx
                jnz     @@450
                inc     ecx
@@460:          mov     [esi],ecx
@@200:          mov     esi,[ebp+8]
                mov     eax,[esi+psign]
                mov     ebx,[ebp+12]
                cmp     eax,[ebx+psign]
                je      @@470
                lea     edi,natz1
                mov     ecx,[edi]
                cmp     ecx,1
                jne     @@480
                cmp     dword ptr [edi+4],0
                je      @@470
@@480:          mov     esi,[ebp+16]
                mov     edx,[esi]
                add     esi,4
                add     edi,4
                sub     edx,ecx
                jecxz   @@490
                lea     esi,[esi+ecx*4]
                lea     edi,[edi+ecx*4]
                neg     ecx
                clc
@@500:          mov     eax,[esi+ecx*4]
                sbb     eax,[edi+ecx*4]
                mov     [edi+ecx*4],eax
                inc     ecx
                jne     @@500
                jnb     @@490
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@510:          mov     eax,[esi+edx*4]
                sub     eax,1
                jnb     @@520
                mov     [edi+edx*4],eax
                inc     edx
                jne     @@510
                jmp     @@530
@@490:          test    edx,edx
                jz      @@530
                lea     esi,[esi+edx*4]
                lea     edi,[edi+edx*4]
                neg     edx
@@540:          mov     eax,[esi+edx*4]
@@520:          mov     [edi+edx*4],eax
                inc     edx
                jne     @@540
@@530:          mov     esi,[ebp+16]
                mov     ecx,[esi]
                lea     edi,natz1
                cmp     dword ptr[edi+ecx*4],0
                jne     @@550
                dec     ecx
                jz      @@560
@@570:          cmp     dword ptr[edi+ecx*4],0
                jne     @@550
                dec     ecx
                jnz     @@570
@@560:          inc     ecx
@@550:          mov     [edi],ecx
@@470:          lea     edi,natz1
                mov     esi,[ebp+20]
                mov     ecx,[edi]
                mov     [esi],ecx
@@580:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@580
                mov     dword ptr [esi+psign],1
end;

procedure zsqrmod (a,b: zin; var c: zin); assembler;
asm
                mov     esi,[ebp+8]
                mov     ecx,[esi]
                lea     ebx,natz3
                mov     dword ptr [ebx+psign],1
                mov     edi,esi
                mov     edx,ecx
                mov     pm,ecx
                add     pm,ecx
                add     esi,4
                add     edi,4
                add     ebx,4
                lea     esi,dword ptr [esi+ecx*4]
                lea     ebx,dword ptr [ebx+ecx*4]
                neg     ecx
                mov     eax,[edi]
                mov     mul0,eax
                mov     loc_lb,edx
                mov     loc_la,ecx
                mov     loc_b,edi
                mov     loc_a,esi
                xor     edi,edi
@@100:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                mov     [ebx+ecx*4],eax
                mov     edi,edx
                inc     ecx
                jne     @@100
                mov     [ebx],edx
                dec     loc_lb
                jz      @@110
@@120:          add     loc_b,4
                mov     edi,loc_b
                mov     edi,[edi]
                mov     mul0,edi
                mov     esi,loc_a
                add     ebx,4
                mov     ecx,loc_la
                xor     edi,edi
@@130:          mov     eax,mul0
                mul     dword ptr [esi+ecx*4]
                add     eax,edi
                adc     edx,0
                add     [ebx+ecx*4],eax
                adc     edx,0
                mov     edi,edx
                inc     ecx
                jne     @@130
                mov     [ebx],edx
                dec     loc_lb
                jnz     @@120
@@110:          lea     edi,natz3
                mov     ecx,pm
@@140:          cmp     dword ptr[edi+ecx*4],0
                jne     @@150
                dec     ecx
                jnz     @@140
                inc     ecx
@@150:          mov     [edi],ecx
@@160:          mov     la1,ecx
                mov     esi,[ebp+12]
                mov     edx,[esi]
                cmp     edx,la1
                jbe     @@170
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@180:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@180
                jmp     @@190
@@170:          mov     la2,edx
                cmp     edx,2
                ja      @@200
                je      @@210
                mov     ebx,[esi+4]
                xor     edx,edx
@@220:          mov     eax,[edi+ecx*4]
                div     ebx
                dec     ecx
                jnz     @@220
                mov     ebx,[ebp+16]
                mov     dword ptr [ebx],1
                mov     [ebx+4],edx
                jmp     @@190
@@210:          mov     ebx,[esi+4]
                mov     eax,[esi+8]
                add     edi,4
                xor     edx,edx
                dec     ecx
                mov     reg_c,ecx
                lea     edi,[edi+ecx*4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                shl     ebx,cl
                mov     reg_b,eax
                mov     reg_a,ebx
                xor     edx,edx
                mov     eax,[edi]
                mov     esi,edx
                sub     edi,4
                mov     ebx,[edi]
                shld    edx,eax,cl
                shld    eax,ebx,cl
@@230:          cmp     edx,reg_b
                je      @@240
                div     reg_b
                jmp     @@250
@@240:          add     edx,eax
                mov     eax,4294967295
                jnc     @@250
                sub     edi,4
                mov     esi,eax
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@260
                mov     ebx,[edi]
@@260:          add     eax,reg_a
                adc     edx,0
                sub     edx,reg_a
                jmp     @@270
@@250:          mov     esi,eax
                sub     edi,4
                mov     eax,ebx
                mov     ebx,reg_c
                dec     ebx
                mov     reg_c,ebx
                pushf
                jz      @@280
                mov     ebx,[edi]
@@280:          shld    eax,ebx,cl
                mov     reg_s,edx
                mov     reg_u,eax
                mov     eax,reg_a
                mul     esi
                sub     eax,reg_u
                sbb     edx,reg_s
                jb      @@290
                ja      @@300
                test    eax,eax
                jz      @@290
@@300:          dec     esi
                sub     eax,reg_a
                sbb     edx,reg_b
                jb      @@290
                ja      @@300
                test    eax,eax
                jnz     @@300
@@290:          neg     eax
                adc     edx,0
                neg     edx
@@270:          popf
                jnz     @@230
                shrd    eax,edx,cl
                shr     edx,cl
                mov     edi,[ebp+16]
                mov     [edi+4],eax
                cmp     edx,0
                jne     @@310
                mov     dword ptr [edi],1
                jmp     @@190
@@310:          mov     [edi+8],edx
                mov     dword ptr [edi],2
                jmp     @@190
@@200:          mov     edx,la1
                mov     esi,[ebp+16]
                mov     [esi],ecx
@@320:          mov     eax,[edi+ecx*4]
                mov     [esi+ecx*4],eax
                dec     ecx
                jnz     @@320
@@330:          lea     esi,[esi+edx*4+4]
                mov     ecx,la2
                sub     edx,ecx
                mov     edi,[ebp+12]
                lea     edi,[edi+ecx*4+4]
                mov     eax,[esi-4]
                cmp     eax,[edi-4]
                jb      @@340
                mov     dword ptr [esi],0
                add     esi,4
                inc     edx
                jmp     @@350
@@340:          test    edx,edx
                jz      @@360
@@350:          mov     ebx,[edi-8]
                mov     eax,[edi-4]
                bsr     ecx,eax
                neg     ecx
                add     ecx,31
                shld    eax,ebx,cl
                mov     reg_a,eax
                mov     eax,[edi-12]
                shld    ebx,eax,cl
                mov     reg_b,ebx
                mov     reg_s,ecx
                mov     reg_c,edx
@@370:          sub     esi,4
                mov     edx,[esi]
                mov     ebx,[esi-4]
                mov     eax,[esi-8]
                mov     ecx,reg_s
                shld    edx,ebx,cl
                shld    ebx,eax,cl
                mov     reg_u,ebx
                mov     ebx,[esi-12]
                shld    eax,ebx,cl
                mov     ebx,eax
                mov     eax,reg_u
                cmp     reg_a,edx
                je      @@380
                div     reg_a
                mov     reg_q,eax
                jmp     @@390
@@380:          add     edx,eax
                mov     reg_q,4294967295
                jc      @@400
@@390:          mov     ecx,edx
                mov     eax,reg_b
                mul     reg_q
                sub     ebx,eax
                sbb     ecx,edx
                jnb     @@400
                dec     reg_q
                add     ebx,reg_b
                adc     ecx,reg_a
                jc      @@400
                dec     reg_q
@@400:          mov     ecx,la2
                neg     ecx
                xor     ebx,ebx
@@410:          mov     eax,[edi+ecx*4]
                mul     reg_q
                add     eax,ebx
                adc     edx,0
                sub     [esi+ecx*4],eax
                adc     edx,0
                mov     ebx,edx
                inc     ecx
                jne     @@410
                cmp     [esi],ebx
                je      @@420
                dec     reg_q
                mov     ecx,la2
                neg     ecx
                clc
@@430:          mov     eax,[edi+ecx*4]
                adc     [esi+ecx*4],eax
                inc     ecx
                jne     @@430
@@420:          dec     reg_c
                jne     @@370
@@360:          mov     esi,[ebp+16]
                mov     ecx,la2
@@440:          cmp     dword ptr[esi+ecx*4],0
                jne     @@450
                dec     ecx
                jnz     @@440
                inc     ecx
@@450:          mov     [esi],ecx
@@190:          mov     esi,[ebp+16]
                mov     dword ptr [esi+psign],1
end;

procedure zmodexp (a,b,c: zin; var d: zin);
begin
nmodexp(a.v,b.v,c.v,d.v);
d.s:=1
end;

procedure zgcd (a,b: zin; var c: zin);
begin
ngcd(a.v,b.v,c.v);
c.s:=1
end;

END.
